import os

import click
import numpy as np
from osteoid import Skeleton

import kimimaro
from kimimaro.utility import mkdir
import fastremap
from tqdm import tqdm

from . import codecs

class Tuple3(click.ParamType):
  """A command line option type consisting of 3 comma-separated integers."""
  name = 'tuple3'
  def convert(self, value, param, ctx):
    if isinstance(value, str):
      try:
        value = tuple(map(int, value.split(',')))
      except ValueError:
        self.fail(f"'{value}' does not contain a comma delimited list of 3 integers.")
      if len(value) != 3:
        self.fail(f"'{value}' does not contain a comma delimited list of 3 integers.")
    return value


@click.group()
def main():
  """
  Skeletonize all labels in a segmented volumetric image
  by applying a TEASAR based algorithm and outputs them
  as SWC.

  Does not accept continuously valued images such as raw
  microscopy images.

  Input File Formats Supported: npy
  
  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version. Run "igneous license" for details.  
  """
  pass

@main.command()
@click.argument("src")
@click.option('--scale', type=float, default=4, help="Adds multiple of boundary distance to invalidation zone. (You should set this!)", show_default=True)
@click.option('--const', type=float, default=10, help="Adds constant physical distance to invalidation zone. (You should set this!)", show_default=True)
@click.option('--pdrf-scale', type=int, default=1e5, help="Constant multiplier of penalty field.", show_default=True)
@click.option('--pdrf-exponent', type=int, default=4, help="Exponent of penalty field. Powers of two are faster. Too big can cause floating point errors.", show_default=True)
@click.option('--soma-detect', type=float, default=750, help="Perform more expensive check for somas for distance to boundary values above this threshold. e.g. 750 nm", show_default=True)
@click.option('--soma-accept', type=float, default=1100, help="Distance to boundary values above this threshold trigger special soma processing. e.g. 750 nm", show_default=True)
@click.option('--soma-scale', type=float, default=2, help="Adds multiple of boundary distance to invalidation zone around a soma. (You should set this!)", show_default=True)
@click.option('--soma-const', type=float, default=300, help="Adds constant physical distance to invalidation zone around a soma. (You should set this!)", show_default=True)
@click.option('--anisotropy', type=Tuple3(), default="1,1,1", help="Physical size of voxel in x,y,z axes.", show_default=True)
@click.option('--dust', type=int, default=1000, help="Skip connected components with fewer voxels than this.", show_default=True)
@click.option('--progress', is_flag=True, default=False, help="Show progress bar.", show_default=True)
@click.option('--fill-holes/--no-fill-holes', is_flag=True, default=True, help="Fill holes in each connected component. (slower)", show_default=True)
@click.option('--fix-avocados', is_flag=True, default=False, help="Use heuristics to combine nucleii with cell bodies. (slower)", show_default=True)
@click.option('--fix-borders', is_flag=True, default=False, help="Center the skeleton where the shape contacts the border.", show_default=True)
@click.option('--fix-branches', is_flag=True, default=True, help="Improves quality of forked shapes. (slower for highly branched shapes)", show_default=True)
@click.option('--max-paths', type=int, default=None, help="Maximum number of paths to trace per object.", show_default=True)
@click.option('-p', '--parallel', type=int, default=1, help="Number of processes to use.", show_default=True)
@click.option('-o', '--outdir', type=str, default="kimimaro_out", help="Where to write the SWC files.", show_default=True)
@click.option("--cross-section", type=int, default=0, help="Turn on cross section analysis. The integer value gives the normal smoothing window, 0=off.", show_default=True)
def forge(
  src,
  scale, const, pdrf_scale, pdrf_exponent,
  soma_detect, soma_accept, soma_scale, soma_const,
  anisotropy, dust, progress, fill_holes, 
  fix_avocados, fix_branches, fix_borders,
  parallel, max_paths, outdir, cross_section,
):
  """Skeletonize an input image and write out SWCs."""
  labels = codecs.load(src)

  skels = kimimaro.skeletonize(
    labels,
    teasar_params={
      "scale": scale,
      "const": const,
      "pdrf_scale": pdrf_scale,
      "pdrf_exponent": pdrf_exponent,
      "soma_detection_threshold": soma_detect,
      "soma_acceptance_threshold": soma_accept,
      "soma_invalidation_scale": soma_scale,
      "soma_invalidation_const": soma_const,
      "max_paths": max_paths,
    },
    anisotropy=anisotropy,
    dust_threshold=dust,
    progress=progress,
    fill_holes=fill_holes,
    fix_avocados=fix_avocados,
    fix_branching=fix_branches,
    fix_borders=fix_borders,
    parallel=parallel,
  )

  directory = mkdir(outdir)

  for label, skel in skels.items():
    fname = os.path.join(directory, f"{label}.swc")
    with open(fname, "wt") as f:
      f.write(skel.to_swc())

  if progress:
    print(f"kimimaro: wrote {len(skels)} skeletons to {directory}")

  if cross_section > 0:
    skels = kimimaro.cross_sectional_area(
      labels, 
      skels,
      anisotropy=anisotropy,
      progress=progress,
      smoothing_window=cross_section,
      fill_holes=fill_holes,
    )

    for label, skel in skels.items():
      fname = os.path.join(directory, f"{label}_xs_area.npy")
      np.save(fname, skel.cross_sectional_area)
      fname = os.path.join(directory, f"{label}_xs_area_contacts.npy")
      np.save(fname, skel.cross_sectional_area_contacts)

    if progress:
      print(f"Wrote cross sectional area and border contacts to {directory}")  

@main.group()
def swc():
  """Utilities for managing SWC files. Use forge to create new skeletons."""
  pass

@swc.command("from")
@click.argument("src", nargs=-1)
def from_image(src):
  """Convert a binary image that has already been skeletonized by a thinning algorithm into an swc."""

  for srcpath in tqdm(src):
    try:
      image = codecs.load(srcpath)
    except ImportError:
      print(f"kimimaro: {srcpath} format not installed.")
      return

    skel = kimimaro.extract_skeleton_from_binary_image(image)

    with open(f"{basename}.swc", "wt") as f:
      f.write(skel.to_swc())

@swc.command("to")
@click.argument("src", nargs=-1)
@click.option('--format', type=str, default="npy", help="Which format to use. Options: npy, tiff", show_default=True)
def to_image(src, format):
  """Convert an swc into a binary image."""
  if format not in ("npy", "tiff"):
    print(f"kimimaro: invalid format {format}. npy or tiff allowed.")

  for srcpath in tqdm(src):
    with open(srcpath, 'rt') as f:
      skel = Skeleton.from_swc(f.read())

    xmin, xmax = fastremap.minmax(skel.vertices[:,0])
    ymin, ymax = fastremap.minmax(skel.vertices[:,1])
    zmin, zmax = fastremap.minmax(skel.vertices[:,2])

    image = np.zeros((int(zmax-zmin), int(ymax-ymin), int(xmax-xmin)), dtype=bool, order='F')
    
    minpt = np.array([int(xmin),int(ymin),int(zmin)])
    drawpts = skel.vertices - minpt
    drawpts = np.asfortranarray(drawpts, dtype=np.int32)
    
    image[np.where((drawpts[:, 0] >= xmin) & (drawpts[:, 0] < xmax) & 
                   (drawpts[:, 1] >= ymin) & (drawpts[:, 1] < ymax) & 
                   (drawpts[:, 2] >= zmin) & (drawpts[:, 2] < zmax))] = True

    basename, ext = os.path.splitext(srcpath)

    if format == "npy":
      np.save(f"{basename}.npy", image)
    elif format == "tiff":
      try:
        import tifffile
        tifffile.imwrite(f"{basename}.tiff", 
                         image.astype(np.float32, copy=False), 
                         photometric='minisblack',
                         metadata={'axes': 'ZYX'},
                         imagej=True)
      except ImportError:
        print("kimimaro: tifffile not installed. Run pip install tifffile.")
        return
    else:
      raise ValueError("should never happen")

@main.command()
@click.argument("filename")
@click.option('--port', type=int, default=8080, help="Which port to run the microviewer on for npy files.", show_default=True)
@click.option('--color-by', type=str, default='r', help="For skeleton visualization. r = radius, c = components, x = cross sectional area (if available).", show_default=True)
def view(filename, port, color_by):
  """Visualize a .swc or .npy file."""
  import microviewer

  basename, ext = os.path.splitext(filename)

  if ext == ".swc":
    with open(filename, "rt") as swc:
      skel = Skeleton.from_swc(swc.read())
    microviewer.objects([ skel ], skeleton_color_by=color_by)
  elif ext == ".npy":
    labels = np.load(filename)
    microviewer.view(labels, seg=True, port=port)
  elif ext == ".ckl":
    import crackle
    labels = crackle.load(filename)
    microviewer.view(labels, seg=True, port=port)
  else:
    print("kimimaro: {filename} was not a .swc, .npy, or .ckl file.")

@main.command()
def license():
  """Prints the license for this library and cli tool."""
  path = os.path.join(os.path.dirname(__file__), 'LICENSE')
  with open(path, 'rt') as f:
    print(f.read())




"""
Kimimaro: TEASAR derived skeletonization for 3D densely labeled images.

Kimimaro is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Kimimaro is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Kimimaro.  If not, see <https://www.gnu.org/licenses/>.
"""

from .intake import skeletonize, DimensionError, synapses_to_targets, connect_points
from .post import postprocess, join_close_components
from .utility import (
	extract_skeleton_from_binary_image,
	cross_sectional_area, 
	cross_sectional_area_single,
	oversegment,
)



"""
Postprocessing for joining skeletons chunks generated by
skeletonizing adjacent image chunks. 

Authors: Alex Bae and Will Silversmith
Affiliation: Seung Lab, Princeton Neuroscience Institue
Date: June 2018 - June 2019

This file is part of Kimimaro.

Kimimaro is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Kimimaro is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Kimimaro.  If not, see <https://www.gnu.org/licenses/>.
"""
from typing import Sequence

from collections import defaultdict

import fastremap
import networkx as nx
import numpy as np

from scipy import spatial
from scipy.sparse import lil_matrix
from scipy.sparse.csgraph import dijkstra
import scipy.sparse.csgraph as csgraph
import scipy.spatial.distance

try:
  from pykdtree.kdtree import KDTree
except ImportError:
  from scipy.spatial import cKDTree as KDTree

from osteoid import Skeleton, Bbox

import kimimaro.skeletontricks

## Public API of Module

def postprocess(
  skeleton:Skeleton, 
  dust_threshold:float = 1500.0, 
  tick_threshold:float = 3000.0,
) -> Skeleton:
  """
  Postprocessing of a skeleton enables aggregation of adjacent
  or overlapping skeletonized image chunks to be fused into a
  single coherent skeleton.  

  The following steps are applied:
  1) Remove disconnected components smaller than the 
      dust threshold (measured in physical distance).
  2) Skeletons are supposed to be trees, so we remove
    any loops that were introduced by joining chunks 
    together. Loops that occur inside the lumen of a 
    neuron might be collapsed into their centroid. Loops
    that occur due to, e.g. mergers are broken arbitarily.
  3) Disconnected components that are closer than the sum
     of their boundary distance are connected.
  4) Small "ticks", or branches from the main skeleton, are
     removed one at a time, from smallest to largest. Branches
     larger than the physical tick_threshold are preserved. 

  Returns: Skeleton
  """
  label = skeleton.id

  # necessary for removing trivial loops etc
  # remove_loops and remove_ticks assume a 
  # clean representation
  skeleton = skeleton.consolidate() 

  skeleton = remove_dust(skeleton, dust_threshold) 
  skeleton = remove_loops(skeleton)
  skeleton = join_close_components(skeleton, restrict_by_radius=True)
  skeleton = remove_ticks(skeleton, tick_threshold)
  skeleton.id = label
  return skeleton.consolidate()

def join_close_components(
  skeletons:Sequence[Skeleton], 
  radius:float = np.inf,
  restrict_by_radius:bool = False,
) -> Skeleton:
  """
  Given a set of skeletons which may contain multiple connected components,
  attempt to connect each component to the nearest other component via the
  nearest two vertices. Repeat until no components remain or no points closer
  than `radius` are available.

  radius: in same units as skeletons, don't join pieces 
    further away than this.
  restrict_by_radius: If the skeletons have a radius property,
    don't join pieces if the neighboring nodes are further away
    than r1 + r2.

  Returns: Skeleton
  """
  if radius is None:
    radius = np.inf

  if radius is not None and radius <= 0:
    raise ValueError("radius must be greater than zero: " + str(radius))

  try:
    iter(skeletons)
  except TypeError:
    skeletons = [ skeletons ]

  skels = []
  for skeleton in skeletons:
    skels += skeleton.components()

  skels = [ skl.consolidate() for skl in skels if not skl.empty() ]

  if len(skels) == 1:
    return skels[0]
  elif len(skels) == 0:
    return Skeleton()

  N = len(skels)
  radii_matrix = np.full( (N, N), np.inf, dtype=np.float32 )
  index_matrix = np.full( (N, N, 2), np.iinfo(np.uint32).max, dtype=np.uint32 )

  if restrict_by_radius:
    radius = 2 * np.max([ np.max(s.radii) for s in skels ])
    radius = max(radius, 0)

  def compute_nearest(tree, i, j):
    s1, s2 = skels[i], skels[j]
    r, idx = tree.query(
      s2.vertices, 
      k=1, 
      distance_upper_bound=(radius + 0.000001), # < bound, so +epsilon
    )
    idx_s2 = np.argmin(r)
    idx_s1 = idx[idx_s2]

    local_radius = r[idx_s2]

    if (
      restrict_by_radius
      and not np.isinf(local_radius)
      and hasattr(s1, "radii")
      and hasattr(s2, "radii")
      and local_radius > (s1.radii[idx_s1] + s2.radii[idx_s2])
    ):
      local_radius = np.inf

    radii_matrix[i,j] = local_radius
    radii_matrix[j,i] = local_radius

    index_matrix[i,j] = ( idx_s1, idx_s2 )
    index_matrix[j,i] = index_matrix[j,i]

  def symmetric_delete(matrix, k):
    matrix = np.delete(matrix, k, axis=0)
    return np.delete(matrix, k, axis=1)

  for i in range(N):
    tree = KDTree(skels[i].vertices)
    for j in range(i + 1, N):  # compute upper triangle only
      compute_nearest(tree, i, j)
    del tree

  while len(skels) > 1:
    
    if np.all(radii_matrix) == np.inf:
      break

    min_radius = np.min(radii_matrix)
    if np.isinf(min_radius) or min_radius > radius:
      break

    i, j = np.unravel_index( np.argmin(radii_matrix), radii_matrix.shape )
    s1, s2 = skels[i], skels[j]
    fused = Skeleton.simple_merge([s1, s2])

    fused.edges = np.concatenate([
      fused.edges,
      [[ index_matrix[i,j,0], index_matrix[i,j,1] + s1.vertices.shape[0] ]]
    ])
    skels[i] = None
    skels[j] = None
    skels = [ fused ] + [ _ for _ in skels if _ is not None ]

    radii_matrix = symmetric_delete(radii_matrix, i)
    radii_matrix = symmetric_delete(radii_matrix, j - 1)
    
    N = len(skels)
    radii_matrix2 = np.full((N,N), np.inf, dtype=np.float32)
    radii_matrix2[1:,1:] = radii_matrix
    radii_matrix = radii_matrix2
    del radii_matrix2

    index_matrix = symmetric_delete(index_matrix, i)
    index_matrix = symmetric_delete(index_matrix, j - 1)
    
    index_matrix2 = np.full((N,N,2), np.iinfo(np.uint32).max, dtype=np.uint32 )
    index_matrix2[1:,1:] = index_matrix
    index_matrix = index_matrix2
    del index_matrix2

    tree = KDTree(skels[0].vertices)
    for j in range(1,N):
      compute_nearest(tree, 0, j)
    del tree

  return Skeleton.simple_merge(skels).consolidate()

## Implementation Details Below

def remove_dust(skeleton, dust_threshold):
  """Dust threshold in physical cable length."""
  
  if skeleton.empty() or dust_threshold == 0:
    return skeleton

  skels = [] 
  for skel in skeleton.components():
    if skel.cable_length() > dust_threshold:
      skels.append(skel)

  return Skeleton.simple_merge(skels)

def remove_ticks(skeleton, threshold):
  """
  Simple merging of individual TESAR cubes results in lots of little 
  ticks due to the edge effect. We can remove them by thresholding
  the path length from a given branch to the "main body" of the neurite. 
  We successively remove paths from shortest to longest until no branches
  below threshold remain.

  If TEASAR parameters were chosen such that they allowed for spines to
  be traced, this is also an opportunity to correct for that.

  This algorithm is O(N^2) in the number of terminal nodes.

  Parameters:
    threshold: The maximum length in nanometers that may be culled.

  Returns: tick free skeleton
  """
  if skeleton.empty() or threshold == 0:
    return skeleton

  skels = []
  for component in skeleton.components():
    skels.append(_remove_ticks(component, threshold))

  return Skeleton.simple_merge(skels).consolidate(remove_disconnected_vertices=False)

def _remove_ticks(skeleton, threshold):
  """
  For a single connected component, remove "ticks" below a threshold. 
  Ticks are a path connecting a terminal node to a branch point that
  are physically shorter than the specified threshold. 

  Every time a tick is removed, it potentially changes the topology
  of the components. Once a branch point's number of edges drops to
  two, the two paths connecting to it can be unified into one. Sometimes
  a single object exists that has no branches but is below threshold. We
  do not delete these objects as there would be nothing left.

  Each time the minimum length tick is removed, it can change which 
  tick is the new minimum tick and requires reevaluation of the whole 
  skeleton. Previously, we did not perform this reevaluation and it 
  resulted in the ends of neurites being clipped. 

  This makes the algorithm quadratic in the number of terminal branches.
  As high resolution skeletons can have tens of thousands of nodes and 
  dozens of branches, a full topological reevaluation becomes relatively 
  expensive. However, we only need to know the graph of distances between
  critical points, defined as the set of branch points and terminal points, 
  in the skeleton in order to evaluate the topology. 

  Therefore, we first compute this distance graph before proceeding with
  tick removal. The algorithm remains quadratic in the number of terminal
  points, but the constant speed up is very large as we move from a regime
  of tens of thousands to hundreds of thousands of points needing reevaluation
  to at most hundreds and often only a handful in typical cases. In the 
  pathological case of a skeleton with numerous single point extrusions,
  the performance of the algorithm collapses approximately to the previous
  regime (though without the assistence of the constant factor of numpy speed).

  Requires:
    skeleton: a Skeleton that is guaranteed to be a single 
      connected component.
    threshold: distance in nanometers below which a branch is considered
      a "tick" eligible to be removed.

  Returns: a "tick" free Skeleton
  """
  if skeleton.empty():
    return skeleton

  dgraph = kimimaro.skeletontricks.create_distance_graph(skeleton)
  vertices = skeleton.vertices
  edges = skeleton.edges

  unique_nodes, unique_counts = fastremap.unique(edges, return_counts=True)
  terminal_nodes = set(unique_nodes[ unique_counts == 1 ])

  branch_idx = np.where(unique_counts >= 3)[0]

  branch_counts = defaultdict(int)
  for i in branch_idx:
    branch_counts[unique_nodes[i]] = unique_counts[i]

  G = nx.Graph()
  G.add_edges_from(edges)

  terminal_superedges = set([ edg for edg in dgraph.keys() if (edg[0] in terminal_nodes or edg[1] in terminal_nodes) ])

  def fuse_edge(edg1):
    unify = [ edg for edg in dgraph.keys() if edg1 in edg ]
    new_dist = 0.0
    for edg in unify:
      terminal_superedges.discard(edg)
      new_dist += dgraph[edg]
      del dgraph[edg]
    unify = set([ item for sublist in unify for item in sublist ])
    unify.remove(edg1)
    dgraph[tuple(unify)] = new_dist
    terminal_superedges.add(tuple(unify))
    branch_counts[edg1] = 0

  while len(dgraph) > 1:
    min_edge = min(terminal_superedges, key=dgraph.get)
    e1, e2 = min_edge

    if branch_counts[e1] == 1 and branch_counts[e2] == 1:
      break
    elif dgraph[min_edge] >= threshold:
      break

    path = nx.shortest_path(G, e1, e2)
    path = [ (path[i], path[i+1]) for i in range(len(path) - 1) ]
    G.remove_edges_from(path)

    del dgraph[min_edge]
    terminal_superedges.remove(min_edge)
    branch_counts[e1] -= 1
    branch_counts[e2] -= 1

    if branch_counts[e1] == 2:
      fuse_edge(e1)
    if branch_counts[e2] == 2:
      fuse_edge(e2)

  skel = skeleton.clone()
  skel.edges = np.array(list(G.edges), dtype=np.uint32)
  return skel

def _create_distance_graph(skeleton):
  """
  Creates the distance "supergraph" from a single connected component 
  skeleton as described in _remove_ticks.

  Returns: a distance "supergraph" describing the physical distance
    between the critical points in the skeleton's structure.

  Example skeleton with output:

      60nm   60nm   60nm     
    1------2------3------4
      30nm |  70nm \
           5        ----6

  { 
    (1,2): 60,  
    (2,3): 60,
    (2,5): 30,
    (3,4): 60,
    (3,6): 70,
  }
  """
  vertices = skeleton.vertices
  edges = skeleton.edges

  unique_nodes, unique_counts = fastremap.unique(edges, return_counts=True)
  terminal_nodes = unique_nodes[ unique_counts == 1 ]
  branch_nodes = set(unique_nodes[ unique_counts >= 3 ])
  
  critical_points = set(terminal_nodes)
  critical_points.update(branch_nodes)

  tree = defaultdict(set)

  for e1, e2 in edges:
    tree[e1].add(e2)
    tree[e2].add(e1)

  # The below depth first search would be
  # more elegantly implemented as recursion,
  # but it quickly blows the stack, mandating
  # an iterative implementation.

  stack = [ terminal_nodes[0] ]
  parents = [ -1 ]
  dist_stack = [ 0.0 ]
  root_stack = [ terminal_nodes[0] ]
  distgraph = defaultdict(float) # the distance "supergraph"

  while stack:
    node = stack.pop()
    dist = dist_stack.pop()
    root = root_stack.pop()
    parent = parents.pop()

    if node in critical_points and node != root:
      distgraph[ (root, node) ] = dist
      dist = 0.0
      root = node

    for child in tree[node]:
      if child != parent:
        stack.append(child)
        parents.append(node)
        dist_stack.append(
          dist + np.linalg.norm(vertices[node,:] - vertices[child,:])
        )
        root_stack.append(root)

  return distgraph

def remove_loops(skeleton):
  if skeleton.empty():
    return skeleton

  skels = []
  for component in skeleton.components():
    skels.append(_remove_loops(component))

  return Skeleton.simple_merge(skels).consolidate(remove_disconnected_vertices=False)

def _remove_loops(skeleton):
  nodes = skeleton.vertices
  edges = np.copy(skeleton.edges).astype(np.int32)

  while True: # Loop until all cycles are removed
    edges = edges.astype(np.int32)
    cycle_path = kimimaro.skeletontricks.find_cycle(edges)
    # cycle_path = kimimaro.skeletontricks.find_cycle_cython(edges)

    if len(cycle_path) == 0:
      break

    edges_cycle = path2edge(cycle_path)

    edges_cycle = np.array(edges_cycle, dtype=np.uint32)
    edges_cycle.sort(axis=1, kind='quicksort')

    nodes_cycle = fastremap.unique(edges_cycle)
    nodes_cycle = nodes_cycle.astype(np.int32)
    
    unique_nodes, unique_counts = fastremap.unique(edges, return_counts=True)
    branch_nodes = unique_nodes[ unique_counts >= 3 ]

    # branch cycles are cycle nodes that coincide with a branch point
    branch_cycle = nodes_cycle[np.isin(nodes_cycle,branch_nodes)]
    branch_cycle = branch_cycle.astype(np.int32)

    # Summary:
    # 0 external branches: isolated loop, just remove it
    # 1 external branch  : remove the loop but draw a line
    #   from the branch point to the farthest node in the loop.
    # 2 external branches: remove the shortest path between
    #   the two entry/exit points. 
    # 3+ external branches: collapse the cycle into its centroid
    #   if the radius of the centroid is less than the EDT radius
    #   of the pixel located at the centroid. Otherwise, arbitrarily
    #   cut an edge from the cycle to break it. This radius rule prevents
    #   issues where we collapse to a point outside of the neurite.

    # Loop with a tail
    if branch_cycle.shape[0] == 1:
      branch_cycle_point = nodes[branch_cycle, :]
      cycle_points = nodes[nodes_cycle, :]

      dist = np.sum((cycle_points - branch_cycle_point) ** 2, 1)
      end_node = nodes_cycle[np.argmax(dist)]

      edges = remove_row(edges, edges_cycle)        
      new_edge = np.array([[branch_cycle[0], end_node]], dtype=np.int32) 
      edges = np.concatenate((edges, new_edge), 0)

    # Loop with an entrance and an exit
    elif branch_cycle.shape[0] == 2:

      # compute the shortest path between the two branch points
      path = np.array(cycle_path[1:])
      pos = np.where(np.isin(path, branch_cycle))[0]
      if (pos[1] - pos[0]) < len(path) / 2:
        path = path[pos[0]:pos[1]+1]
      else:
        path = np.concatenate((path[pos[1]:], path[:pos[0]+1]), 0)

      edge_path = path2edge(path)
      edge_path.sort(axis=1, kind='quicksort')

      row_valid = np.ones(edges_cycle.shape[0])
      for i in range(edge_path.shape[0]):
        row_valid -= (edges_cycle[:,0] == edge_path[i,0]) * (edges_cycle[:,1] == edge_path[i,1])

      row_valid = row_valid.astype(bool)
      edge_path = edges_cycle[row_valid,:]

      edges = remove_row(edges, edge_path)

    # Totally isolated loop
    elif branch_cycle.shape[0] == 0:
      edges = remove_row(edges, edges_cycle)

    # Loops with many ways in and out
    # looks like here we unify them into their
    # centroid. This doesn't work well if the loop
    # is large.
    else:
      branch_cycle_points = nodes[branch_cycle,:]

      centroid = np.mean(branch_cycle_points, axis=0)
      dist = (nodes - centroid)
      dist *= dist
      dist = np.sum(dist, axis=1)
      intersect_node = np.argmin(dist)
      intersect_point = nodes[intersect_node,:]

      dist = np.sum((branch_cycle_points - intersect_point) ** 2, 1)
      dist = np.sqrt(np.max(dist))

      # Fix the "stargate" issue where a large loop
      # can join lots of things to the near center
      # by just making a tiny snip if the distance
      # is greater than the radius of the connected node.
      if dist > skeleton.radii[ intersect_node ]:
        edges = remove_row(edges, edges_cycle[:1,:])
        continue

      edges = remove_row(edges, edges_cycle)      

      new_edges = np.zeros((branch_cycle.shape[0], 2))
      new_edges[:,0] = branch_cycle
      new_edges[:,1] = intersect_node

      if np.isin(intersect_node, branch_cycle):
        idx = np.where(branch_cycle == intersect_node)
        new_edges = np.delete(new_edges, idx, 0)

      edges = np.concatenate((edges,new_edges), 0)

  skeleton.vertices = nodes
  skeleton.edges = edges.astype(np.uint32)
  return skeleton

def path2edge(path):
  """
  path: sequence of nodes

  Returns: sequence separated into edges
  """
  edges = np.zeros([len(path) - 1, 2], dtype=np.uint32)
  edges[:,0] = path[0:-1]
  edges[:,1] = path[1:]
  return edges

def remove_row(array, rows2remove): 
  array.sort(axis=1, kind='quicksort')
  if array.size == 0:
    return array.astype(np.int32, copy=False)

  rows2remove.sort(axis=1, kind='quicksort')

  for i in range(rows2remove.shape[0]):  
    idx = find_row(array,rows2remove[i,:])  
    if np.sum(idx == -1) == 0: 
      array = np.delete(array, idx, axis=0) 
  
  return array.astype(np.int32, copy=False)

def find_row(array, row): 
  """ 
  array: array to search for  
  row: row to find  
   Returns: row indices 
  """ 
  matches = (array[:,0] == row[0])
  matches &= (array[:,1] == row[1])
  idx = np.where(matches)
  if len(idx) == 0:
    return -1
  return idx[0]



"""
Skeletonization algorithm based on TEASAR (Sato et al. 2000).

Authors: Alex Bae and Will Silversmith
Affiliation: Seung Lab, Princeton Neuroscience Institue
Date: June 2018 - Februrary 2025

This file is part of Kimimaro.

Kimimaro is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Kimimaro is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Kimimaro.  If not, see <https://www.gnu.org/licenses/>.
"""
from collections import defaultdict
from math import log

import dijkstra3d
import edt
import fill_voids
import numpy as np
from scipy import ndimage

import kimimaro.skeletontricks

from osteoid import Skeleton

def trace(
    labels, DBF, 
    scale=10, const=10, anisotropy=(1,1,1), 
    soma_detection_threshold=1100, 
    soma_acceptance_threshold=4000, 
    pdrf_scale=5000, pdrf_exponent=16,
    soma_invalidation_scale=0.5,
    soma_invalidation_const=0,
    fix_branching=True,
    manual_targets_before=[],
    manual_targets_after=[],
    root=None,
    max_paths=None,
    voxel_graph=None,
  ):
  """
  Given the euclidean distance transform of a label ("Distance to Boundary Function"), 
  convert it into a skeleton using an algorithm based on TEASAR. 

  DBF: Result of the euclidean distance transform. Must represent a single label,
       assumed to be expressed in chosen physical units (i.e. nm)
  scale: during the "rolling ball" invalidation phase, multiply the DBF value by this.
  const: during the "rolling ball" invalidation phase, this is the minimum radius in chosen physical units (i.e. nm).
  anisotropy: (x,y,z) conversion factor for voxels to chosen physical units (i.e. nm)
  soma_detection_threshold: if object has a DBF value larger than this, 
    root will be placed at largest DBF value and special one time invalidation
    will be run over that root location (see soma_invalidation scale)
    expressed in chosen physical units (i.e. nm) 
  pdrf_scale: scale factor in front of dbf, used to weight dbf over euclidean distance (higher to pay more attention to dbf) (default 5000)
  pdrf_exponent: exponent in dbf formula on distance from edge, faster if factor of 2 (default 16)
  soma_invalidation_scale: the 'scale' factor used in the one time soma root invalidation (default .5)
  soma_invalidation_const: the 'const' factor used in the one time soma root invalidation (default 0)
                           (units in chosen physical units (i.e. nm))
  fix_branching: When enabled, zero out the graph edge weights traversed by 
    of previously found paths. This causes branch points to occur closer to 
    the actual path divergence. However, there is a large performance penalty
    associated with this as dijkstra's algorithm is computed once per a path
    rather than once per a skeleton.
  manual_targets_before: list of (x,y,z) that correspond to locations that must 
    have paths drawn to. Used for specifying root and border targets for
    merging adjacent chunks out-of-core. Targets are applied before ordinary
    target selection.
  manual_targets_after: Same as manual_targets_before but the additional 
    targets are applied after the usual algorithm runs. The current 
    invalidation status of the shape makes no difference.
  max_paths: If a label requires drawing this number of paths or more,
    abort and move onto the next label.
  root: If you want to force the root to be a particular voxel, you can
    specify it here.
  voxel_graph: a connection graph that defines permissible 
    directions of motion between voxels. This is useful for
    dealing with self-touches. The graph is defined by the
    conventions used in cc3d.voxel_connectivity_graph 
    (https://github.com/seung-lab/connected-components-3d/blob/3.2.0/cc3d_graphs.hpp#L73-L92)

  Based on the algorithm by:

  M. Sato, I. Bitter, M. Bender, A. Kaufman, and M. Nakajima. 
  "TEASAR: tree-structure extraction algorithm for accurate and robust skeletons"  
    Proc. the Eighth Pacific Conference on Computer Graphics and Applications. Oct. 2000.
    doi:10.1109/PCCGA.2000.883951 (https://ieeexplore.ieee.org/document/883951/)

  Returns: Skeleton object
  """
  dbf_max = np.max(DBF)
  labels = np.asfortranarray(labels)
  DBF = np.asfortranarray(DBF)

  soma_mode = False
  # > 5000 nm, gonna be a soma or blood vessel
  # For somata: specially handle the root by 
  # placing it at the approximate center of the soma
  if dbf_max > soma_detection_threshold:
    labels, num_voxels_filled = fill_voids.fill(labels, in_place=True, return_fill_count=True)
    if num_voxels_filled > 0:
      del DBF
      DBF = edt.edt(
        labels, 
        anisotropy=anisotropy, 
        black_border=np.all(labels),
        voxel_graph=voxel_graph,
      )
    dbf_max = np.max(DBF) 
    soma_mode = dbf_max > soma_acceptance_threshold

  soma_radius = 0.0

  if soma_mode:
    if root is not None:
      manual_targets_before.insert(0, root)
    root = find_soma_root(DBF, dbf_max)    
    soma_radius = dbf_max * soma_invalidation_scale + soma_invalidation_const
  elif root is None:
    root = find_root(labels, anisotropy, voxel_graph)
  
  if root is None:
    return Skeleton()
 
  free_space_radius = 0 if not soma_mode else DBF[root]
  # DBF: Distance to Boundary Field
  # DAF: Distance from any voxel Field (distance from root field)
  # PDRF: Penalized Distance from Root Field
  DBF = kimimaro.skeletontricks.zero2inf(DBF) # DBF[ DBF == 0 ] = np.inf
  DAF, target = dijkstra3d.euclidean_distance_field(
    labels, root, 
    anisotropy=anisotropy, 
    free_space_radius=free_space_radius,
    voxel_graph=voxel_graph,
    return_max_location=True,
  )
  DAF = kimimaro.skeletontricks.inf2zero(DAF) # DAF[ DAF == np.inf ] = 0
  target_finder = kimimaro.skeletontricks.CachedTargetFinder(labels, DAF)
  PDRF = compute_pdrf(dbf_max, pdrf_scale, pdrf_exponent, DBF, DAF, DAF[target])
  del DAF

  # Use dijkstra propogation w/o a target to generate a field of
  # pointers from each voxel to its parent. Then we can rapidly
  # compute multiple paths by simply hopping pointers using path_from_parents
  if not fix_branching:
    parents = dijkstra3d.parental_field(PDRF, root, voxel_graph=voxel_graph)
    del PDRF
  else:
    parents = PDRF

  if soma_mode:
    invalidated, labels = kimimaro.skeletontricks.roll_invalidation_ball_inside_component(
      labels, DBF, 
      soma_invalidation_scale,
      soma_invalidation_const, 
      anisotropy,
      [root],
      voxel_connectivity_graph=voxel_graph,
    )
  # This target is only valid if no 
  # invalidations have occured yet.
  elif len(manual_targets_before) == 0:
    manual_targets_before.append(target)
  
  paths = compute_paths(
    root, labels, DBF, target_finder, 
    parents, scale, const, anisotropy, 
    soma_mode, soma_radius, fix_branching,
    manual_targets_before, manual_targets_after, 
    max_paths, voxel_graph
  )

  skel = Skeleton.simple_merge(
    [ Skeleton.from_path(path) for path in paths if len(path) > 0 ]
  ).consolidate()

  verts = skel.vertices.flatten().astype(np.uint32)
  skel.radii = DBF[verts[::3], verts[1::3], verts[2::3]]
  skel.transform = np.array([
    [anisotropy[0], 0, 0, 0],
    [0, anisotropy[1], 0, 0],
    [0, 0, anisotropy[2], 0],
  ], dtype=np.float32)

  return skel

def compute_paths(
    root, labels, DBF, target_finder, 
    parents, scale, const, anisotropy, 
    soma_mode, soma_radius, fix_branching,
    manual_targets_before, manual_targets_after,
    max_paths, voxel_graph
  ):
  """
  Given the labels, DBF, DAF, dijkstra parents,
  and associated invalidation knobs, find the set of paths 
  that cover the object. Somas are given special treatment
  in that we attempt to cull vertices within a radius of the
  root vertex.
  """
  paths = []
  valid_labels = np.count_nonzero(labels)
  root = tuple(root)

  if max_paths is None:
    max_paths = valid_labels

  if len(manual_targets_before) + len(manual_targets_after) >= max_paths:
    return []

  parents[tuple(root)] = 0 # provide initial rail for dijkstra.railroad

  while (valid_labels > 0 or manual_targets_before or manual_targets_after) \
    and len(paths) < max_paths:

    if manual_targets_before:
      target = manual_targets_before.pop()
    elif valid_labels == 0:
      target = manual_targets_after.pop()
    else:
      target = target_finder.find_target(labels)

    if fix_branching:
      # Draw a path (a "road") from the target to the nearest zero weighted
      # path (a "rail"). This has some minor efficiencies vs drawing
      # from a target all the way to the source. Also, target -> source
      # is much more efficient than source -> target for three reasons.
      # (a) target -> catches a rail instead of exploring all rails
      # (b) target has a natural edge effect that restrict exploration
      # (c) in soma, target -> source follows gradients vs fights them
      path = dijkstra3d.railroad(
        parents, target, voxel_graph=voxel_graph
      )
    else:
      path = dijkstra3d.path_from_parents(parents, target)
    
    if soma_mode:
      dist_to_soma_root = np.linalg.norm(anisotropy * (path - root), axis=1)
      # remove all path points which are within soma_radius of root
      path = np.concatenate(
        (path[:1,:], path[dist_to_soma_root > soma_radius, :])
      )

    if valid_labels > 0:
      invalidated, labels = kimimaro.skeletontricks.roll_invalidation_ball_inside_component(
        labels, DBF, scale, const, 
        anisotropy, path,
        voxel_connectivity_graph=voxel_graph,
      )      
      valid_labels -= invalidated

    for vertex in path:
      if fix_branching:
        parents[tuple(vertex)] = 0.0

    paths.append(path)

  return paths

def find_soma_root(DBF, dbf_max):
  """
  This perhaps overcomplicates things, but it's possible,
  for example in a rectangular cuboid, for there to be
  many multiple maxima at the center of a shape. We pick
  the one closest to the centroid of the shape to ensure
  the choice is sensible.

  Returns: (x,y,z) as integers
  """
  maxima = (DBF == dbf_max)
  com = ndimage.measurements.center_of_mass(maxima)
  com = np.asarray(com, dtype=np.float32)
  
  coords = np.where(maxima)
  coords = np.vstack( coords ).T
  root = np.argmin(
    np.sum((coords - com) ** 2, axis=1)
  )

  return tuple(coords[root].astype(np.uint32))

def find_root(labels, anisotropy, voxel_graph):
  """
  "4.4 DAF:  Compute distance from any voxel field"
  Compute DAF, but we immediately convert to the PDRF
  The extremal point of the PDRF is a valid root node
  even if the DAF is computed from an arbitrary pixel.
  """
  any_voxel = kimimaro.skeletontricks.first_label(labels)   
  if any_voxel is None: 
    return None

  DAF, target = dijkstra3d.euclidean_distance_field(
    labels, any_voxel, 
    anisotropy=anisotropy,
    return_max_location=True,
    voxel_graph=voxel_graph,
  )
  return target

def is_power_of_two(num):
  if int(num) != num:
    return False
  return num != 0 and ((num & (num - 1)) == 0)

def compute_pdrf(
  dbf_max, pdrf_scale, 
  pdrf_exponent, DBF, DAF,
  max_daf
):
  """
  Add p(v) to the DAF (pp. 4, section 4.5)
  "4.5 PDRF: Compute penalized distance from root voxel field"
  Let M > max(DBF)
  p(v) = 5000 * (1 - DBF(v) / M)^16
  5000 is chosen to allow skeleton segments to be up to 3000 voxels
  long without exceeding floating point precision.

  IMPLEMENTATION NOTE: 
  Appearently repeated *= is much faster than "** f(16)" 
  12,740.0 microseconds vs 4 x 560 = 2,240 microseconds (5.69x)

  More clearly written:
  PDRF = DAF + 5000 * ((1 - DBF * M) ** 16)
  """
  f = lambda x: np.float32(x)
  M = f( 1 / (dbf_max ** 1.01) )

  # First branch is much faster than ** which presumably
  # uses logarithms to do the exponentiation.
  PDRF = np.empty(DBF.shape, dtype=np.float32, order="F")
  np.multiply(DBF, M, out=PDRF)
  np.subtract(f(1), PDRF, out=PDRF)
  if is_power_of_two(pdrf_exponent) and (pdrf_exponent < (2 ** 16)):
    for _ in range(int(np.log2(pdrf_exponent))):
      PDRF *= PDRF # ^pdrf_exponent
  else: 
    np.power(PDRF, pdrf_exponent, out=PDRF)

  PDRF *= f(pdrf_scale)

  # provide trickle of gradient so open spaces don't collapse
  if max_daf != 0:
    DAF *= (1 / max_daf)
    PDRF += DAF

  return np.asfortranarray(PDRF)

def point_to_point(
  binary_img, start, end,
  anisotropy=(1,1,1), 
  pdrf_scale=100000, 
  pdrf_exponent=4,
):
  """
  Trace a single centerline path from 
  start to end.
  """
  DBF = edt.edt(
    binary_img, 
    anisotropy=anisotropy,
    black_border=True,
  )
  dbf_max = np.max(DBF)

  DBF = kimimaro.skeletontricks.zero2inf(DBF) # DBF[ DBF == 0 ] = np.inf
  DAF, target = dijkstra3d.euclidean_distance_field(
    binary_img, start, 
    anisotropy=anisotropy,
    return_max_location=True,
  )
  DAF = kimimaro.skeletontricks.inf2zero(DAF) # DAF[ DAF == np.inf ] = 0
  PDRF = compute_pdrf(dbf_max, pdrf_scale, pdrf_exponent, DBF, DAF, DAF[target])
  del DAF

  path = dijkstra3d.dijkstra(PDRF, end, start)
  skel = Skeleton.from_path(path)

  verts = skel.vertices.flatten().astype(np.uint32)
  skel.radii = DBF[verts[::3], verts[1::3], verts[2::3]]
  return skel



"""
This file is part of Kimimaro.

Kimimaro is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Kimimaro is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Kimimaro.  If not, see <https://www.gnu.org/licenses/>.
"""

from collections import defaultdict
from functools import partial
import gc
import multiprocessing as mp
import signal
import uuid

import numpy as np
import pathos.pools
import scipy.spatial
from tqdm import tqdm

from osteoid import Skeleton, Bbox

import cc3d # connected components
from crackle import CrackleArray
import edt # euclidean distance transform
import fastremap
import fill_voids

import kimimaro.skeletontricks
import kimimaro.trace

from . import sharedmemory as shm
from .utility import compute_cc_labels, find_objects

class DimensionError(Exception):
  pass

DEFAULT_TEASAR_PARAMS = {
  "scale": 1.5, 
  "const": 300,
  "pdrf_scale": 100000,
  "pdrf_exponent": 4,
  "soma_acceptance_threshold": 3500,
  "soma_detection_threshold": 750,
  "soma_invalidation_const": 300,
  "soma_invalidation_scale": 2
}

def skeletonize(
  all_labels, teasar_params=DEFAULT_TEASAR_PARAMS, anisotropy=(1,1,1),
  object_ids=None, dust_threshold=1000, 
  progress=True, fix_branching=True, in_place=False, 
  fix_borders=True, parallel=1, parallel_chunk_size=100,
  extra_targets_before=[], extra_targets_after=[],
  fill_holes=False, fix_avocados=False,
  voxel_graph=None
):
  """
  Skeletonize all non-zero labels in a given 2D or 3D image.

  Required:
    all_labels: a 2D or 3D numpy array of integer type (signed or unsigned) 

  Optional:
    anisotropy: the physical dimensions of each axis (e.g. 4nm x 4nm x 40nm)
    object_ids: If not none, zero out all labels other than those specified here.
    teasar_params: {
      scale: during the "rolling ball" invalidation phase, multiply 
          the DBF value by this.
      const: during the "rolling ball" invalidation phase, this 
          is the minimum radius in chosen physical units (i.e. nm).
      soma_detection_threshold: if object has a DBF value larger than this, 
          root will be placed at largest DBF value and special one time invalidation
          will be run over that root location (see soma_invalidation scale)
          expressed in chosen physical units (i.e. nm) 
      pdrf_scale: scale factor in front of dbf, used to weight dbf over euclidean distance (higher to pay more attention to dbf) (default 5000)
      pdrf_exponent: exponent in dbf formula on distance from edge, faster if factor of 2 (default 16)
      soma_invalidation_scale: the 'scale' factor used in the one time soma root invalidation (default .5)
      soma_invalidation_const: the 'const' factor used in the one time soma root invalidation (default 0)
                             (units in chosen physical units (i.e. nm))
      max_paths: max paths to trace on a single object. Moves onto the next object after this point.
    }
    dust_threshold: don't bother skeletonizing connected components smaller than
      this many voxels.
    fill_holes: preemptively run a void filling algorithm on all connected
      components and delete labels that get filled in. This can improve the
      quality of the reconstruction if holes in the shapes are artifacts introduced
      by the segmentation pipeline. This option incurs moderate overhead.

      WARNING: THIS WILL REMOVE INPUT LABELS THAT ARE DEEMED TO BE HOLES.

    extra_targets_before: List of x,y,z voxel coordinates that will all 
      be traced to from the root regardless of whether those points have 
      been invalidated. These targets will be applied BEFORE the regular
      target selection algorithm is run.      

      e.g. [ (x,y,z), (x,y,z) ]

    extra_targets_after: Same as extra_targets_before but the additional
      targets will be applied AFTER the usual algorithm runs.

    progress: if true, display a progress bar
    fix_branching: When enabled, zero the edge weights by of previously 
      traced paths. This causes branch points to occur closer to 
      the actual path divergence. However, there is a performance penalty
      associated with this as dijkstra's algorithm is computed once per a path
      rather than once per a skeleton.
    in_place: if true, allow input labels to be modified to reduce
      memory usage and possibly improve performance.
    fix_borders: ensure that segments touching the border place a 
      skeleton endpoint in a predictable place to make merging 
      adjacent chunks easier.
    fix_avocados: If nuclei are segmented seperately from somata
      then we can try to detect and fix this issue.
    voxel_graph: a connection graph that defines permissible 
      directions of motion between voxels. This is useful for
      dealing with self-touches. The graph is defined by the
      conventions used in cc3d.voxel_connectivity_graph 
      (https://github.com/seung-lab/connected-components-3d/blob/3.2.0/cc3d_graphs.hpp#L73-L92)
    parallel: number of subprocesses to use.
      <= 0: Use multiprocessing.count_cpu() 
         1: Only use the main process.
      >= 2: Use this number of subprocesses.
    parallel_chunk_size: default number of skeletons to 
      submit to each parallel process before returning results,
      updating the progress bar, and submitting a new task set. 
      Setting this number too low results in excess IPC overhead,
      and setting it too high can result in task starvation towards
      the end of a job and infrequent progress bar updates. If the
      chunk size is set higher than num tasks // parallel, that number
      is used instead.

  Returns: { $segid: osteoid.Skeleton, ... }
  """

  anisotropy = np.array(anisotropy, dtype=np.float32)

  all_labels = format_labels(all_labels, in_place=in_place)
  all_labels = apply_object_mask(all_labels, object_ids)

  if all_labels.size <= dust_threshold:
    return {}
  
  if isinstance(all_labels, CrackleArray):
    minlabel = all_labels.min()
    maxlabel = all_labels.max()
  else:
    minlabel, maxlabel = fastremap.minmax(all_labels)

  if minlabel == 0 and maxlabel == 0:
    return {}

  cc_labels, remapping = compute_cc_labels(all_labels, voxel_graph)
  del all_labels

  if isinstance(cc_labels, CrackleArray) and (fill_holes or voxel_graph or fix_avocados):
    cc_labels = cc_labels.numpy()

  if fill_holes:
    cc_labels = fill_all_holes(cc_labels, progress)

  extra_targets_before = points_to_labels(extra_targets_before, cc_labels)
  extra_targets_after = points_to_labels(extra_targets_after, cc_labels)

  def edtfn(labels):
    if isinstance(labels, CrackleArray):
      labels = labels[:]

    return edt.edt(labels, 
      anisotropy=anisotropy,
      black_border=(minlabel == maxlabel),
      parallel=parallel,
      voxel_graph=voxel_graph,
    )

  all_dbf = edtfn(cc_labels)
  
  if fix_avocados:
    cc_labels, all_dbf, remapping = engage_avocado_protection(
      cc_labels, all_dbf, remapping,
      soma_detection_threshold=teasar_params.get('soma_detection_threshold', 0),
      edtfn=edtfn,
      progress=progress,
    )

  if isinstance(cc_labels, CrackleArray):
    cc_ct_iterator = cc_labels.voxel_counts().items()
  else:
    cc_segids, pxct = fastremap.unique(cc_labels, return_counts=True)
    cc_ct_iterator = zip(cc_segids, pxct)
  
  cc_segids = [ sid for sid, ct in cc_ct_iterator if ct > dust_threshold and sid != 0 ]

  all_slices = find_objects(cc_labels)

  border_targets = defaultdict(list)
  if fix_borders:
    border_targets = compute_border_targets(cc_labels, anisotropy)

  print_quotes(parallel) # easter egg

  if parallel <= 0:
    parallel = mp.cpu_count()

  if parallel == 1:
    return skeletonize_subset(
      all_dbf, cc_labels, voxel_graph, remapping, 
      teasar_params, anisotropy, all_slices, 
      border_targets, extra_targets_before, extra_targets_after,
      progress, fix_borders, fix_branching, 
      cc_segids
    )
  else:
    # The following section can't be moved into 
    # skeletonize parallel because then all_dbf 
    # and cc_labels can't be deleted to save memory.
    suffix = uuid.uuid1().hex

    dbf_shm_location = 'kimimaro-shm-dbf-' + suffix
    cc_shm_location = 'kimimaro-shm-cc-labels-' + suffix
    vg_shm_location = 'kimimaro-shm-voxel-graph-' + suffix

    try:
      dbf_mmap, all_dbf_shm = shm.ndarray( all_dbf.shape, all_dbf.dtype, dbf_shm_location, order='F')
      all_dbf_shm[:] = all_dbf 
      del all_dbf 

      cc_mmap, cc_labels_shm = shm.ndarray( cc_labels.shape, cc_labels.dtype, cc_shm_location, order='F')    
      cc_labels_shm[:] = cc_labels[:]
      del cc_labels

      voxel_graph_shm = None
      vg_mmap = None
      if voxel_graph is not None:
        vg_mmap, voxel_graph_shm = shm.ndarray( voxel_graph.shape, voxel_graph.dtype, vg_shm_location, order='F')    
        voxel_graph_shm[:] = voxel_graph
        del voxel_graph

      skeletons = skeletonize_parallel(      
        all_dbf_shm, dbf_shm_location, 
        cc_labels_shm, cc_shm_location, remapping, 
        voxel_graph_shm, vg_shm_location,
        teasar_params, anisotropy, all_slices, 
        border_targets, extra_targets_before, extra_targets_after,
        progress, fix_borders, fix_branching, 
        cc_segids, parallel, parallel_chunk_size
      )
    finally:
      dbf_mmap.close()
      cc_mmap.close()
      shm.unlink(dbf_shm_location)
      shm.unlink(cc_shm_location)
      if voxel_graph is not None:
        vg_mmap.close()
        shm.unlink(vg_shm_location)

    return skeletons

def connect_points(
  labels, start, end,
  anisotropy=(1,1,1), 
  fill_holes=False, 
  in_place=False,
  pdrf_scale=100000, 
  pdrf_exponent=4,
):
  """
  Extract a single centerline skeleton between
  two preselected points from a binary image.

  labels: a 2D or 3D binary image
  start: an (x,y,z) tuple
  end: an (x,y,z) tuple

  anisotropy: the physical dimensions of each axis (e.g. 4nm x 4nm x 40nm)
  fill_holes: preemptively run a void filling algorithm on all connected
    components and delete labels that get filled in. This can improve the
    quality of the reconstruction if holes in the shapes are artifacts introduced
    by the segmentation pipeline.

  pdrf_scale: scale factor in front of dbf, used to weight dbf over euclidean distance (higher to pay more attention to dbf)
  pdrf_exponent: exponent in dbf formula on distance from edge, faster if factor of 2
  """
  anisotropy = np.array(anisotropy, dtype=np.float32)
  start = tuple(start)
  end = tuple(end)

  labels = labels.astype(np.bool)
  labels = format_labels(labels, in_place=in_place)

  cc_labels, remapping = compute_cc_labels(labels)
  if cc_labels[start] == 0 or cc_labels[start] != cc_labels[end]:
    raise ValueError("Cannot extract centerline from disconnected components.")
  del cc_labels
  del remapping

  skel = kimimaro.trace.point_to_point(
    labels, start, end,
    anisotropy=anisotropy, 
    pdrf_scale=pdrf_scale, pdrf_exponent=pdrf_exponent,
  )
  skel.vertices *= anisotropy
  skel.space = 'physical'
  return skel

def format_labels(labels, in_place):
  if isinstance(labels, CrackleArray):
    return labels

  if in_place:
    labels = fastremap.asfortranarray(labels)
  else:
    labels = np.copy(labels, order='F')

  if labels.dtype == bool:
    labels = labels.view(np.uint8)

  original_shape = labels.shape

  while labels.ndim < 3:
    labels = labels[..., np.newaxis ]

  while labels.ndim > 3:
    if labels.shape[-1] == 1:
      labels = labels[..., 0]
    else:
      raise DimensionError(
        "Input labels may be no more than three non-trivial dimensions. Got: {}".format(
          original_shape
        )
      )

  return labels

def skeletonize_parallel(
    all_dbf_shm, dbf_shm_location, 
    cc_labels_shm, cc_shm_location, remapping, 
    voxel_graph_shm, vg_shm_location,
    teasar_params, anisotropy, all_slices, 
    border_targets, extra_targets_before, extra_targets_after,
    progress, fix_borders, fix_branching, 
    cc_segids, parallel, chunk_size
  ):
    prevsigint = signal.getsignal(signal.SIGINT)
    prevsigterm = signal.getsignal(signal.SIGTERM)
    
    # Don't fork, spawn entirely new processes. This
    # avoids accidental deadlocks.
    mp.set_start_method("spawn", force=True)
    
    executor = pathos.pools.ProcessPool(parallel)

    def cleanup(signum, frame):
      shm.unlink(dbf_shm_location)
      shm.unlink(cc_shm_location)
      executor.terminate()

    signal.signal(signal.SIGINT, cleanup)
    signal.signal(signal.SIGTERM, cleanup)   

    vg_shape = voxel_graph_shm.shape if voxel_graph_shm else None
    vg_dtype = voxel_graph_shm.dtype if voxel_graph_shm else None

    skeletonizefn = partial(parallel_skeletonize_subset, 
      dbf_shm_location, all_dbf_shm.shape, all_dbf_shm.dtype, 
      cc_shm_location, cc_labels_shm.shape, cc_labels_shm.dtype,
      vg_shm_location, vg_shape, vg_dtype,
      remapping, teasar_params, anisotropy, all_slices, 
      border_targets, extra_targets_before, extra_targets_after, 
      False, # progress, use our own progress bar below
      fix_borders, fix_branching, 
    )

    ccids = []
    if chunk_size < len(cc_segids) // parallel:
      for i in range(0, len(cc_segids), chunk_size):
        ccids.append(cc_segids[i:i+chunk_size])
    else:
      for i in range(parallel):
        ccids.append(cc_segids[i::parallel])

    skeletons = defaultdict(list)
    with tqdm(total=len(cc_segids), disable=(not progress), desc="Skeletonizing Labels") as pbar:
      for skels in executor.uimap(skeletonizefn, ccids):
        for segid, skel in skels.items():
          skeletons[segid].append(skel)
        pbar.update(len(skels))
    executor.close()
    executor.join()
    executor.clear()

    signal.signal(signal.SIGINT, prevsigint)
    signal.signal(signal.SIGTERM, prevsigterm)
    
    shm.unlink(dbf_shm_location)
    shm.unlink(cc_shm_location)
    shm.unlink(vg_shm_location)

    return merge(skeletons)

def parallel_skeletonize_subset(    
    dbf_shm_location, dbf_shape, dbf_dtype, 
    cc_shm_location, cc_shape, cc_dtype, 
    vg_shm_location, vg_shape, vg_dtype,
    *args, **kwargs
  ):
  
  dbf_mmap, all_dbf = shm.ndarray( dbf_shape, dtype=dbf_dtype, location=dbf_shm_location, order='F')
  cc_mmap, cc_labels = shm.ndarray( cc_shape, dtype=cc_dtype, location=cc_shm_location, order='F')

  if vg_shape is None:
    vg_mmap, voxel_graph = None, None
  else:
    vg_mmap, voxel_graph = shm.ndarray( vg_shape, dtype=vg_dtype, location=vg_shm_location, order='F')

  skels = skeletonize_subset(all_dbf, cc_labels, voxel_graph, *args, **kwargs)

  dbf_mmap.close()
  cc_mmap.close()
  if vg_mmap:
    vg_mmap.close()

  return skels

def skeletonize_subset(
    all_dbf, cc_labels, voxel_graph, remapping, 
    teasar_params, anisotropy, all_slices, 
    border_targets, extra_targets_before, extra_targets_after,
    progress, fix_borders, fix_branching, 
    cc_segids
  ):

  skeletons = defaultdict(list)

  with tqdm(cc_segids, disable=(not progress), desc="Skeletonizing Labels") as pbar:
    for segid in pbar:

      pbar.set_postfix(label=str(remapping[segid]))

      # Crop DBF to ROI
      slices = all_slices[segid - 1]
      if slices is None:
        continue

      roi = Bbox.from_slices(slices)
      if roi.volume() <= 1:
        continue

      if isinstance(cc_labels, CrackleArray):
        labels = cc_labels.decompress(label=segid, crop=True)
        label_slcs = (slices[0], slices[1], slice(None))
        labels = np.asfortranarray(labels[label_slcs])
      else:
        labels = cc_labels[slices]
        labels = (labels == segid)

      dbf = np.where(labels, all_dbf[slices], 0.0)
      cropped_voxel_graph = (voxel_graph[slices] if voxel_graph is not None else None)

      manual_targets_before = []
      manual_targets_after = []
      root = None 

      def translate_to_roi(targets):
        targets = np.array(targets)
        targets -= roi.minpt.astype(np.uint32)
        return targets.tolist()      

      # We only source a predetermined root from 
      # border_targets because we understand that it's
      # located at a reasonable place at the edge of the
      # shape. In theory, extra targets can be positioned
      # anywhere within the shape or off the shape, making it 
      # a dicey proposition. 
      if len(border_targets[segid]) > 0:
        manual_targets_before = translate_to_roi(border_targets[segid])
        root = manual_targets_before.pop()

      if segid in extra_targets_before and len(extra_targets_before[segid]) > 0:
        manual_targets_before.extend( translate_to_roi(extra_targets_before[segid]) )

      if segid in extra_targets_after and len(extra_targets_after[segid]) > 0:
        manual_targets_after.extend( translate_to_roi(extra_targets_after[segid]) )

      skeleton = kimimaro.trace.trace(
        labels, 
        dbf, 
        anisotropy=anisotropy, 
        fix_branching=fix_branching, 
        manual_targets_before=manual_targets_before,
        manual_targets_after=manual_targets_after,
        root=root,
        voxel_graph=cropped_voxel_graph,
        **teasar_params
      )

      if skeleton.empty():
        continue

      skeleton.vertices += roi.minpt

      orig_segid = remapping[segid]
      skeleton.id = orig_segid
      skeleton.vertices *= anisotropy
      skeleton.space = 'physical'
      skeletons[orig_segid].append(skeleton)

  return merge(skeletons)

def apply_object_mask(all_labels, object_ids):
  if object_ids is None:
    return all_labels

  if isinstance(all_labels, CrackleArray):
    mask = all_labels.labels()
    mask = { u: 0 for u in mask }
    for segid in object_ids:
      mask[segid] = segid
    return all_labels.remap(mask).condense()

  if len(object_ids) == 1:
    all_labels = kimimaro.skeletontricks.zero_out_all_except(all_labels, object_ids[0]) # faster
  else:
    all_labels = fastremap.mask_except(all_labels, object_ids, in_place=True)

  return all_labels

def points_to_labels(pts, cc_labels):
  mapping = defaultdict(list)
  for pt in pts:
    pt = tuple(pt)
    mapping[ cc_labels[pt] ].append(pt)
  return mapping

def compute_border_targets(cc_labels, anisotropy):
  sx, sy, sz = cc_labels.shape

  if isinstance(cc_labels, CrackleArray):
    cc_labels = cc_labels.numpy()

  planes = (
    ( cc_labels[:,:,0], (0, 1), lambda x,y: (x, y, 0) ),     # top xy
    ( cc_labels[:,:,-1], (0, 1), lambda x,y: (x, y, sz-1) ), # bottom xy
    ( cc_labels[:,0,:], (0, 2), lambda x,z: (x, 0, z) ),     # left xz
    ( cc_labels[:,-1,:], (0, 2), lambda x,z: (x, sy-1, z) ), # right xz
    ( cc_labels[0,:,:], (1, 2), lambda y,z: (0, y, z) ),     # front yz
    ( cc_labels[-1,:,:], (1, 2), lambda y,z: (sx-1, y, z) )  # back yz
  )

  target_list = defaultdict(set)

  for plane, dims, rotatefn in planes:
    wx, wy = anisotropy[dims[0]], anisotropy[dims[1]]
    plane = np.copy(plane, order='F')
    cc_plane = cc3d.connected_components(np.ascontiguousarray(plane))
    dt_plane = edt.edt(cc_plane, black_border=True, anisotropy=(wx, wy))

    plane_targets = kimimaro.skeletontricks.find_border_targets(
      dt_plane, cc_plane, wx, wy
    )

    plane = plane[..., np.newaxis]
    cc_plane = cc_plane[..., np.newaxis]
    remapping = kimimaro.skeletontricks.get_mapping(plane, cc_plane)

    for label, pt in plane_targets.items():
      label = remapping[label]
      target_list[label].add(
        rotatefn( int(pt[0]), int(pt[1]) )
      )

  target_list.default_factory = lambda: np.array([], np.uint32)
  for label, pts in target_list.items():
    target_list[label] = np.array(list(pts), dtype=np.uint32)

  return target_list

def merge(skeletons):
  merged_skels = {}
  for segid, skels in skeletons.items():
    skel = Skeleton.simple_merge(skels)
    merged_skels[segid] = skel.consolidate()

  return merged_skels

def argmax(arr):
  if arr.flags['C_CONTIGUOUS']:
    return np.unravel_index(np.argmax(arr), arr.shape, order='C')
  return np.unravel_index(np.argmax(arr.T), arr.shape, order='F')

def engage_avocado_protection(
  cc_labels, all_dbf, remapping,
  soma_detection_threshold, edtfn, 
  progress
):
  orig_cc_labels = np.copy(cc_labels, order='F')

  unchanged = set()

  # This loop handles nested avocados
  # Unless there are deeply nested double avocados,
  # this should complete in 2-3 passes. We limit it
  # to 20 just to make sure this loop terminates no matter what.
  # Avocados aren't the end of the world.
  for _ in tqdm(range(20), disable=(not progress), desc="Avocado Pass"): 
    # Note: Divide soma_detection_threshold by a bit more than 2 because the nucleii are going to be
    # about a factor of 2 or less smaller than what we'd expect from a cell. For example,
    # in an avocado I saw, the DBF of the nucleus was 499 when the detection threshold was 
    # set to 1100.
    candidates = set(fastremap.unique(cc_labels * (all_dbf > soma_detection_threshold / 2.5)))
    candidates -= unchanged
    candidates.discard(0)

    cc_labels, unchanged_this_cycle, changes = engage_avocado_protection_single_pass(
      cc_labels, all_dbf,
      candidates=candidates,
      progress=progress,
    )
    unchanged |= unchanged_this_cycle

    if len(changes) == 0:
      break 
    
    all_dbf = edtfn(cc_labels)

  # Downstream logic assumes cc_labels is contigiously numbered
  cc_labels, _ = fastremap.renumber(cc_labels, in_place=True)
  cc_remapping = kimimaro.skeletontricks.get_mapping(orig_cc_labels, cc_labels)

  adjusted_remapping = {}
  for new_cc, cc in cc_remapping.items():
    if cc in remapping:
      adjusted_remapping[new_cc] = remapping[cc]

  return cc_labels, all_dbf, adjusted_remapping

def engage_avocado_protection_single_pass(
  cc_labels, all_dbf, 
  candidates=None, progress=False
):
  """
  For each candidate, check if there's a fruit around the
  avocado pit roughly from the center (the max EDT).
  """

  if candidates is None:
    candidates = fastremap.unique(cc_labels)

  candidates = [ label for label in candidates if label != 0 ]

  unchanged = set()
  changed = set()

  if len(candidates) == 0:
    return cc_labels, unchanged, changed

  def paint_walls(binimg):
    """
    Ensure that inclusions that touch the wall are handled
    by performing a 2D fill on each wall.
    """
    binimg[:,:,0 ] = fill_voids.fill(binimg[:,:,0 ])
    binimg[:,:,-1] = fill_voids.fill(binimg[:,:,-1])
    binimg[:,0,: ] = fill_voids.fill(binimg[:,0,: ])
    binimg[:,-1,:] = fill_voids.fill(binimg[:,-1,:])
    binimg[0,:,: ] = fill_voids.fill(binimg[0,:,: ])
    binimg[-1,:,:] = fill_voids.fill(binimg[-1,:,:])
    return binimg

  slcs = find_objects(cc_labels)

  for label in tqdm(candidates, disable=(not progress), desc="Fixing Avocados"):
    slc = slcs[label - 1]
    offset = Bbox.from_slices(slc).minpt
    binimg = paint_walls(cc_labels[slc] == label) # image of the pit
    coord = argmax(binimg * all_dbf[slc]) + offset

    (pit, fruit) = kimimaro.skeletontricks.find_avocado_fruit(
      cc_labels, coord[0], coord[1], coord[2]
    )
    if pit == fruit and pit not in changed:
      unchanged.add(pit)
    else:
      unchanged.discard(pit)
      unchanged.discard(fruit)
      changed.add(pit)
      changed.add(fruit)
      binimg |= (cc_labels[slc] == fruit)

    fruit = np.asarray(fruit, dtype=cc_labels.dtype)
    binimg, N = fill_voids.fill(binimg, in_place=True, return_fill_count=True)
    cc_labels[slc] *= ~binimg
    cc_labels[slc] += fruit * binimg

  return cc_labels, unchanged, changed

def synapses_to_targets(labels, synapses, progress=False):
  """
  Turn the output of synapse detection and assignment, usually 
  centroid + pre/post into actionable targets. For a given 
  labeled volume, take the centroid and a pre or post label
  and find the nearest voxel for that label and add the coordinates
  of that voxel to a list of targets.

  labels: a 3d array containing labels
  synapses: { label: [ (centroid, swc_label), (centroid, swc_label), ... ] }
    where centroid is an (x,y,z) float triple in voxel coordinate space
      where the origin is the same as for labels
    where swc_label is the label to be added to the vertex attributes for
      the resulting target.
    where label is a presynaptic OR a postsynaptic label
      (submit two items to cover both)

  Returns: { (x,y,z): swc_label, ... } targets for skeletonization
  """
  while labels.ndim > 3:
    labels = labels[...,0]

  targets = {}

  for label, pairs in tqdm(synapses.items(), disable=(not progress), desc='Converting Synapses to Targets'):
    point_cloud = np.vstack((labels == label).nonzero()).T # [ [x,y,z], ... ]
    if len(point_cloud) == 0:
      continue

    swc_labels = defaultdict(list) 
    for centroid, swc_label in pairs:
      swc_labels[swc_label].append(centroid)

    for swc_label, centroids in swc_labels.items():
      distances = scipy.spatial.distance.cdist(point_cloud, centroids)
      minima = np.unique(np.argmin(distances, axis=0))
      tmp_targets = [ tuple(point_cloud[idx]) for idx in minima ]
      targets.update({ target: swc_label for target in tmp_targets })

  return targets

def fill_all_holes(cc_labels, progress=False, return_fill_count=False):
  """
  Fills the holes in each connected component and removes components that
  get filled in. The idea is that holes (entirely contained labels or background) 
  are artifacts in cell segmentations. A common example is a nucleus segmented 
  separately from the rest of the cell or errors in a manual segmentation leaving
  a void in a dendrite.

  cc_labels: an image containing connected components with labels smaller than
    the number of voxels in the image.
  progress: Display a progress bar or not.
  return_fill_count: if specified, return a tuple (filled_image, N) where N is
    the number of voxels that were filled in.

  Returns: filled_in_labels
  """
  labels = fastremap.unique(cc_labels)
  labels_set = set(labels)
  labels_set.discard(0)

  all_slices = find_objects(cc_labels)
  pixels_filled = 0

  for label in tqdm(labels, disable=(not progress), desc="Filling Holes"):
    if label not in labels_set:
      continue

    slices = all_slices[label - 1]
    if slices is None:
      continue

    binary_image = (cc_labels[slices] == label)
    binary_image, N = fill_voids.fill(
      binary_image, in_place=True, 
      return_fill_count=True
    )
    pixels_filled += N
    if N == 0:
      continue 

    sub_labels = set(fastremap.unique(cc_labels[slices] * binary_image))
    sub_labels.remove(label)
    labels_set -= sub_labels
    cc_labels[slices] = cc_labels[slices] * ~binary_image + label * binary_image

  if return_fill_count:
    return cc_labels, pixels_filled
  return cc_labels

def print_quotes(parallel):
  if parallel == -1:
    print("Against the power of will I possess... The capability of my body is nothing.")
  elif parallel == -2:
    print("I will see the truth of this world... OROCHIMARU-SAMA WILL SHOW ME!!!")

  if -2 <= parallel < 0:
    print("CURSED SEAL OF THE EARTH!!!")  



from collections import defaultdict
import errno
import mmap
import os
import sys
import time

import multiprocessing as mp

import numpy as np

from osteoid import Bbox, Vec

from .utility import mkdir

SHM_DIRECTORY = '/dev/shm/'
EMULATED_SHM_DIRECTORY = '/tmp/kimimaro-shm'

EMULATE_SHM = not os.path.isdir(SHM_DIRECTORY)
PLATFORM_SHM_DIRECTORY = SHM_DIRECTORY if not EMULATE_SHM else EMULATED_SHM_DIRECTORY

class SharedMemoryReadError(Exception):
  pass

class SharedMemoryAllocationError(Exception):
  pass

def ndarray(shape, dtype, location, order='F', readonly=False, lock=None, **kwargs):
  """
  Create a shared memory numpy array. 
  Lock is only necessary while doing multiprocessing on 
  platforms without /dev/shm type  shared memory as 
  filesystem emulation will be used instead.

  Allocating the shared array requires cleanup on your part.
  A shared memory file will be located at sharedmemory.PLATFORM_SHM_DIRECTORY + location
  and must be unlinked when you're done. It will outlive the program.

  You should also call .close() on the mmap file handle when done. However,
  this is less of a problem because the operating system will close the
  file handle on process termination.

  Parameters:
  shape: same as numpy.ndarray
  dtype: same as numpy.ndarray
  location: the shared memory filename 
  lock: (optional) multiprocessing.Lock

  Returns: (mmap filehandle, shared ndarray)
  """
  if EMULATE_SHM:
    return ndarray_fs(
      shape, dtype, location, lock, 
      readonly, order, emulate_shm=True, **kwargs
    )
  return ndarray_shm(shape, dtype, location, readonly, order, **kwargs)

def ndarray_fs(
    shape, dtype, location, lock, 
    readonly=False, order='F', emulate_shm=False,
    **kwargs
  ):
  """Emulate shared memory using the filesystem."""
  dbytes = np.dtype(dtype).itemsize
  nbytes = Vec(*shape).rectVolume() * dbytes

  if emulate_shm:
    directory = mkdir(EMULATED_SHM_DIRECTORY)
    filename = os.path.join(directory, location)
  else:
    filename = location

  if lock:
    lock.acquire()

  try:
    allocate_shm_file(filename, nbytes, dbytes, readonly)
  finally:
    if lock:
      lock.release()

  with open(filename, 'r+b') as f:
    array_like = mmap.mmap(f.fileno(), 0) # map entire file
  
  renderbuffer = np.ndarray(buffer=array_like, dtype=dtype, shape=shape, order=order, **kwargs)
  renderbuffer.setflags(write=(not readonly))
  return array_like, renderbuffer

def allocate_shm_file(filename, nbytes, dbytes, readonly):
  try:
    size = os.path.getsize(filename)
    exists = True
  except FileNotFoundError:
    size = 0
    exists = False

  if readonly and not exists:
    raise SharedMemoryReadError(filename + " has not been allocated. Requested " + str(nbytes) + " bytes.")
  elif readonly and size != nbytes:
    raise SharedMemoryReadError("{} exists, but the allocation size ({} bytes) does not match the request ({} bytes).".format(
      filename, size, nbytes
    ))

  if exists: 
    if size > nbytes:
      with open(filename, 'wb') as f:
        os.ftruncate(f.fileno(), nbytes)
    elif size < nbytes:
      # too small? just remake it below
      os.unlink(filename) 

  exists = os.path.exists(filename)

  if not exists:
    # Previously we were writing out real files full of zeros, 
    # but a) that takes forever and b) modern OSes support sparse
    # files (i.e. gigabytes of zeros that take up only a few real bytes).
    #
    # The following should take advantage of this functionality and be faster.
    # It should work on Python 2.7 Unix, and Python 3.5+ on Unix and Windows.
    #
    # References:
    #   https://stackoverflow.com/questions/8816059/create-file-of-particular-size-in-python
    #   https://docs.python.org/3/library/os.html#os.ftruncate
    #   https://docs.python.org/2/library/os.html#os.ftruncate
    #
    with open(filename, 'wb') as f:
      os.ftruncate(f.fileno(), nbytes)

def ndarray_shm(shape, dtype, location, readonly=False, order='F', **kwargs):
  """Create a shared memory numpy array. Requires /dev/shm to exist."""
  import posix_ipc
  from posix_ipc import O_CREAT
  import psutil

  nbytes = Vec(*shape).rectVolume() * np.dtype(dtype).itemsize
  available = psutil.virtual_memory().available

  preexisting = 0
  # This might only work on Ubuntu
  shmloc = os.path.join(SHM_DIRECTORY, location)
  if os.path.exists(shmloc):
    preexisting = os.path.getsize(shmloc)
  elif readonly:
    raise SharedMemoryReadError(shmloc + " has not been allocated. Requested " + str(nbytes) + " bytes.")

  if readonly and preexisting != nbytes:
    raise SharedMemoryReadError("{} exists, but the allocation size ({} bytes) does not match the request ({} bytes).".format(
      shmloc, preexisting, nbytes
    ))

  if (nbytes - preexisting) > available:
    overallocated = nbytes - preexisting - available
    overpercent = (100 * overallocated / (preexisting + available))
    raise SharedMemoryAllocationError("""
      Requested more memory than is available. 

      Shared Memory Location:  {}

      Shape:                   {}
      Requested Bytes:         {} 
      
      Available Bytes:         {} 
      Preexisting Bytes*:      {} 

      Overallocated Bytes*:    {} (+{:.2f}%)

      * Preexisting is only correct on linux systems that support /dev/shm/""" \
        .format(location, shape, nbytes, available, preexisting, overallocated, overpercent))

  # This might seem like we're being "extra safe" but consider
  # a threading condition where the condition of the shared memory
  # was adjusted between the check above and now. Better to make sure
  # that we don't accidently change anything if readonly is set.
  flags = 0 if readonly else O_CREAT 
  size = 0 if readonly else int(nbytes) 

  try:
    shared = posix_ipc.SharedMemory(location, flags=flags, size=size)
    array_like = mmap.mmap(shared.fd, shared.size)
    os.close(shared.fd)
    renderbuffer = np.ndarray(buffer=array_like, dtype=dtype, shape=shape, order=order, **kwargs)
  except OSError as err:
    if err.errno == errno.ENOMEM: # Out of Memory
      posix_ipc.unlink_shared_memory(location)      
    raise

  renderbuffer.setflags(write=(not readonly))
  return array_like, renderbuffer

def unlink(location):
  if EMULATE_SHM:
    return unlink_fs(location)
  return unlink_shm(location)

def unlink_shm(location):
  import posix_ipc
  try:
    posix_ipc.unlink_shared_memory(location)
  except posix_ipc.ExistentialError:
    return False
  return True

def unlink_fs(location):
  directory = mkdir(EMULATED_SHM_DIRECTORY)
  try:
    filename = os.path.join(directory, location)
    os.unlink(filename)
    return True
  except OSError:
    return False



import numpy as np
import os
import gzip

def normalize_file_ext(filename):
  filename, ext = os.path.splitext(filename)

  two_pass = ('.ckl', '.cpso')

  if ext in two_pass:
    return ext

  while True:
    filename, ext2 = os.path.splitext(filename)
    if ext2 in two_pass:
      return ext2
    elif ext2 == '':
      return ext
    ext = ext2

def load(filename):
  ext = normalize_file_ext(filename)

  if ext == ".ckl":
    import crackle
    return crackle.aload(filename)
  elif ext == ".npy":
    if filename.endswith(".gz"):
      with gzip.GzipFile(filename, "rb") as f:
        image = np.load(f)
    else:
      image = np.load(filename)
  elif ext == ".nrrd":
    import nrrd
    image, header = nrrd.read(filename)
    if image.shape[0] == 3 and image.ndim == 3:
      image = image[...,np.newaxis]
      image = np.transpose(image, axes=[1,2,3,0])
    return image
  elif ext == ".nii":
    import nibabel as nib
    image = nib.load(filename)
    image = np.array(image.dataobj)
  elif ext in (".tif", ".tiff"):
    import tifffile
    image = tifffile.imread(srcpath)
  else:
    raise ValueError("Data type not supported: " + ext)

  return np.asfortranarray(image)



# cython: language_level=3
"""
Certain operations have to be fast for the skeletonization
procedure. The ones that didn't fit elsewhere have a home here.

Author: William Silversmith
Affiliation: Seung Lab, Princeton Neuroscience Institute
Date: August 2018 - May 2024

*****************************************************************
This file is part of Kimimaro.

Kimimaro is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Kimimaro is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Kimimaro.  If not, see <https://www.gnu.org/licenses/>.
*****************************************************************
"""
cimport cython
from libc.stdlib cimport calloc, free
from libc.stdint cimport (
  int8_t, int16_t, int32_t, int64_t,
  uint8_t, uint16_t, uint32_t, uint64_t
)
from libcpp cimport bool
from cpython cimport array 
import array
import sys

from libcpp.vector cimport vector
from libcpp.unordered_map cimport unordered_map
from libcpp.unordered_set cimport unordered_set
from libcpp.utility cimport pair as cpp_pair

cimport numpy as cnp
import numpy as np

cnp.import_array()

from collections import defaultdict

cdef float INFINITY = float('inf')

ctypedef fused UINT:
  uint8_t
  uint16_t
  uint32_t
  uint64_t
  unsigned char

ctypedef fused INTEGER: 
  int8_t
  int16_t
  int32_t
  int64_t
  UINT

cdef extern from "dijkstra_invalidation.hpp" namespace "dijkstra_invalidation":
  cdef int64_t _roll_invalidation_ball(
    uint8_t* field,
    uint64_t sx, uint64_t sy, uint64_t sz, 
    float wx, float wy, float wz, 
    vector[uint64_t] sources,
    vector[float] max_distances,
    int connectivity,
    uint32_t* voxel_connectivity_graph
  )

cdef extern from "skeletontricks.hpp" namespace "skeletontricks":
  cdef size_t _roll_invalidation_cube(
    uint8_t* labels, float* DBF,
    int64_t sx, int64_t sy, int64_t sz,
    float wx, float wy, float wz,
    size_t* path, size_t path_size,
    float scale, float constant
  )

  cdef vector[T] _find_cycle[T](T* edges, size_t Ne)
  
  cdef unordered_map[ uint64_t, float ] _create_distance_graph(
    float* vertices, size_t Nv, 
    uint32_t* edges, size_t Ne, uint32_t start_node,
    vector[int32_t] critical_points_vec
  )

  cdef struct pair_hash:
    size_t __call__(cpp_pair[uint64_t,uint64_t] v)
  cdef unordered_set[ cpp_pair[uint64_t, uint64_t], pair_hash ] _extract_edges_from_binary_image(
    uint8_t* image, 
    uint64_t sx, uint64_t sy, uint64_t sz,
    int connectivity
  )

def find_cycle(cnp.ndarray[int32_t, ndim=2] edges):
  """
  Given a graph of edges that are a single connected component,
  find a cycle via depth first search.

  Returns: list of edges in a cycle (empty list if no cycle is found)
  """
  if edges.size == 0:
    return np.zeros((0,), dtype=np.uint32)

  edges = np.ascontiguousarray(edges)

  cdef cnp.ndarray[int32_t, ndim=1] elist = np.array(
    _find_cycle[int32_t](
      <int32_t*>&edges[0,0], <size_t>(edges.size // 2)
    ),
    dtype=np.int32
  )
  return elist

def create_distance_graph(skeleton):
  """
  Creates the distance "supergraph" from a single connected component 
  skeleton as described in _remove_ticks.

  Returns: a distance "supergraph" describing the physical distance
    between the critical points in the skeleton's structure.

  Example skeleton with output:

      60nm   60nm   60nm     
    1------2------3------4
      30nm |  70nm \
           5        ----6

  { 
    (1,2): 60,  
    (2,3): 60,
    (2,5): 30,
    (3,4): 60,
    (3,6): 70,
  }
  """
  cdef cnp.ndarray[float, ndim=2] vertices = skeleton.vertices
  cdef cnp.ndarray[uint32_t, ndim=2] edges = skeleton.edges

  unique_nodes, unique_counts = np.unique(edges, return_counts=True)
  terminal_nodes = unique_nodes[ unique_counts == 1 ]
  branch_nodes = set(unique_nodes[ unique_counts >= 3 ])
  
  critical_points = set(terminal_nodes)
  critical_points.update(branch_nodes)

  res = _create_distance_graph(
    <float*>&vertices[0,0], vertices.shape[0],
    <uint32_t*>&edges[0,0], edges.shape[0], terminal_nodes[0],
    list(critical_points)
  )
  cdef dict supergraph = res

  cdef dict real_supergraph = {}
  cdef uint64_t key = 0
  cdef int32_t e1, e2

  for key in supergraph.keys():
    e2 = <int32_t>(key & 0xffffffff)
    e1 = <int32_t>(key >> 32)
    real_supergraph[ (e1, e2) ] = supergraph[key]

  return real_supergraph


@cython.boundscheck(False)
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
def inf2zero(cnp.ndarray[float, cast=True, ndim=3] field):
  """
  inf2zero(cnp.ndarray[float, cast=True, ndim=3] field)

  Convert infinities to zeros.

  Returns: field
  """
  cdef size_t sx, sy, sz 
  cdef size_t  x,  y,  z

  sx = field.shape[0]
  sy = field.shape[1]
  sz = field.shape[2]

  for z in range(0, sz):
    for y in range(0, sy):
      for x in range(0, sx):
        if field[x,y,z] == INFINITY:
          field[x,y,z] = 0

  return field

@cython.boundscheck(False)
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
def zero2inf(cnp.ndarray[float, cast=True, ndim=3] field):
  """
  zero2inf(cnp.ndarray[float, cast=True, ndim=3] field)

  Convert zeros to positive infinities.

  Returns: field
  """
  cdef size_t sx, sy, sz 
  cdef size_t  x,  y,  z

  sx = field.shape[0]
  sy = field.shape[1]
  sz = field.shape[2]

  for z in range(0, sz):
    for y in range(0, sy):
      for x in range(0, sx):
        if (field[x,y,z] == 0):
          field[x,y,z] = INFINITY

  return field

@cython.boundscheck(False)  
@cython.wraparound(False)  # turn off negative index wrapping for entire function 
@cython.nonecheck(False)  
def zero_out_all_except(cnp.ndarray[INTEGER, cast=True, ndim=3] field, INTEGER leave_alone): 
  """
  zero_out_all_except(cnp.ndarray[INTEGER, cast=True, ndim=3] field, INTEGER leave_alone)

  Change all values in field to zero except `leave_alone`.

  Returns: field
  """
  cdef size_t sx, sy, sz   
  cdef size_t  x,  y,  z 

  sx = field.shape[0]  
  sy = field.shape[1] 
  sz = field.shape[2] 

  for z in range(0, sz): 
    for y in range(0, sy):  
      for x in range(0, sx):  
        if (field[x,y,z] != leave_alone): 
          field[x,y,z] = 0  

  return field  

@cython.boundscheck(False)
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
def finite_max(cnp.ndarray[float, cast=True, ndim=3] field):
  """
  float finite_max(cnp.ndarray[float, cast=True, ndim=3] field)

  Given a field of floats that may include infinities, find the 
  largest finite value.
  """
  cdef size_t sx, sy, sz 
  cdef size_t  x,  y,  z

  sx = field.shape[0]
  sy = field.shape[1]
  sz = field.shape[2]

  cdef float maximum = -INFINITY
  for z in range(0, sz):
    for y in range(0, sy):
      for x in range(0, sx):
        if (field[x,y,z] > maximum) and (field[x,y,z] < +INFINITY):
          maximum = field[x,y,z]

  return maximum

@cython.boundscheck(False)
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
def finite_min(cnp.ndarray[float, cast=True, ndim=3] field):
  """
  float finite_min(cnp.ndarray[float, cast=True, ndim=3] field)

  Given a field of floats that may include infinities, find the 
  minimum finite value.
  """
  cdef size_t sx, sy, sz 
  cdef size_t  x,  y,  z

  sx = field.shape[0]
  sy = field.shape[1]
  sz = field.shape[2]

  cdef float minimum = -INFINITY
  for z in range(0, sz):
    for y in range(0, sy):
      for x in range(0, sx):
        if (field[x,y,z] < minimum) and (field[x,y,z] > -INFINITY):
          minimum = field[x,y,z]

  return minimum

@cython.boundscheck(False)
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
def first_label(cnp.ndarray[uint8_t, cast=True, ndim=3] labels):
  """
  uint8_t first_label(cnp.ndarray[uint8_t, cast=True, ndim=3] labels)

  Scan through labels to find the first non-zero value and return it.
  """
  cdef size_t sx, sy, sz 
  cdef size_t  x,  y,  z

  sx = labels.shape[0]
  sy = labels.shape[1]
  sz = labels.shape[2]

  for z in range(0, sz):
    for y in range(0, sy):
      for x in range(0, sx):
        if labels[x,y,z]:
          return (x,y,z)

  return None

@cython.boundscheck(False)
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
def find_target(
    cnp.ndarray[uint8_t, cast=True, ndim=3] labels, 
    cnp.ndarray[float, ndim=3] PDRF
  ):
  """
  find_target(ndarray[uint8_t, cast=True, ndim=3] labels, ndarray[float, ndim=3] PDRF)

  Given a binary image and a coregistered map of values to it, 
  find the coordinate of the voxel corresponding to the first
  instance of the maximum map value.

  Returns: (x, y, z)
  """
  cdef size_t x,y,z
  cdef size_t sx, sy, sz

  sx = labels.shape[0]
  sy = labels.shape[1]
  sz = labels.shape[2]

  cdef int64_t mx, my, mz

  mx = -1
  my = -1
  mz = -1

  cdef float maxpdrf = -INFINITY
  for x in range(0, sx):
    for y in range(0, sy):
      for z in range(0, sz):
        if labels[x,y,z] and PDRF[x,y,z] > maxpdrf:
          maxpdrf = PDRF[x,y,z]
          mx = x
          my = y
          mz = z

  return (mx, my, mz)

@cython.boundscheck(False)
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.binding(True)
def roll_invalidation_ball_inside_component(
    cnp.ndarray[uint8_t, cast=True, ndim=3] labels, 
    cnp.ndarray[float, ndim=3] DBF, 
    float scale, 
    float constant,
    anisotropy,
    path,
    voxel_connectivity_graph = None,
    connectivity = 26,
):
  cdef int64_t sx, sy, sz 
  sx = labels.shape[0]
  sy = labels.shape[1]
  sz = labels.shape[2]

  cdef size_t sxy = sx * sy

  cdef float wx, wy, wz
  (wx, wy, wz) = anisotropy

  max_distances = [ 
    (scale * DBF[x,y,z] + constant) for (x,y,z) in path 
  ]

  path = [ 
    coord[0] + sx * coord[1] + sxy * coord[2] 
    for coord in path if tuple(coord)
  ]

  cdef uint32_t* vcg = NULL
  cdef cnp.ndarray[uint32_t, ndim=3] vcg_arr

  if isinstance(voxel_connectivity_graph, np.ndarray):
    vcg_arr = voxel_connectivity_graph
    vcg = <uint32_t*>&vcg_arr[0,0,0]

  invalidated = _roll_invalidation_ball(
    <uint8_t*>&labels[0,0,0],
    sx, sy, sz, 
    wx, wy, wz,
    path, max_distances,
    connectivity, 
    vcg
  )

  return (invalidated, labels)

@cython.boundscheck(False)
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.binding(True)
def roll_invalidation_ball(
    cnp.ndarray[uint8_t, cast=True, ndim=3] labels, 
    cnp.ndarray[float, ndim=3] DBF, 
    path, float scale, float const,
    anisotropy=(1,1,1),
    invalid_vertices={},
  ):
  """
  Given an anisotropic binary image, its distance transform, and a path 
  traversing the binary image, erase the voxels surrounding the path
  in a sphere around each vertex on the path corresponding to the 
  equation: 

  r = scale * DBF[x,y,z] + const

  Returns: modified labels
  """
  cdef int64_t sx, sy, sz 
  sx = labels.shape[0]
  sy = labels.shape[1]
  sz = labels.shape[2]

  cdef float wx, wy, wz
  (wx, wy, wz) = anisotropy
    
  cdef float radius, dist
  cdef int64_t minx, maxx, miny, maxy, minz, maxz

  cdef int64_t x,y,z
  cdef int64_t x0, y0, z0

  cdef size_t invalidated = 0

  for coord in path:
    if tuple(coord) in invalid_vertices:
      continue

    (x0, y0, z0) = coord
    radius = DBF[x0,y0,z0] * scale + const # physical units (e.g. nm)

    minx = max(0,  <int64_t>(0.5 + (x0 - (radius / wx))))
    maxx = min(sx, <int64_t>(0.5 + (x0 + (radius / wx))))
    miny = max(0,  <int64_t>(0.5 + (y0 - (radius / wy))))
    maxy = min(sy, <int64_t>(0.5 + (y0 + (radius / wy))))
    minz = max(0,  <int64_t>(0.5 + (z0 - (radius / wz))))
    maxz = min(sz, <int64_t>(0.5 + (z0 + (radius / wz))))

    radius *= radius 

    for x in range(minx, maxx):
      for y in range(miny, maxy):
        for z in range(minz, maxz):
          if not labels[x,y,z]:
            continue 

          dist = (wx * (x - x0)) ** 2 + (wy * (y - y0)) ** 2 + (wz * (z - z0)) ** 2
          if dist <= radius:
            invalidated += 1
            labels[x,y,z] = 0

  return invalidated, labels

@cython.boundscheck(False)
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.binding(True)
def get_mapping(
    cnp.ndarray[INTEGER, ndim=3] orig_labels, 
    cnp.ndarray[UINT, ndim=3] cc_labels
  ):
  """
  Given a set of possibly not connected labels 
  and an image containing their labeled connected components, 
  produce a dictionary containing the inverse of this mapping.

  Returns: { $CC_LABEL: $ORIGINAL_LABEL }
  """

  cdef size_t sx, sy, sz 
  sx = orig_labels.shape[0]
  sy = orig_labels.shape[1]
  sz = orig_labels.shape[2]

  cdef size_t x,y,z 

  remap = {}

  if orig_labels.size == 0:
    return remap

  cdef UINT last_label = cc_labels[0,0,0]
  remap[cc_labels[0,0,0]] = orig_labels[0,0,0]

  for z in range(sz):
    for y in range(sy):
      for x in range(sx):
        if last_label == cc_labels[x,y,z]:
          continue
        remap[cc_labels[x,y,z]] = orig_labels[x,y,z]
        last_label = cc_labels[x,y,z]

  return remap

@cython.binding(True)
def compute_centroids(
    cnp.ndarray[UINT, ndim=2] labels,
    float wx, float wy
  ):
  """
  Compute the centroid for every label on a 2D image at once.

  Returns: { $segid: (x, y), ... }
  """

  cdef float[:] xsum = np.zeros( (labels.size,), dtype=np.float32)
  cdef float[:] ysum = np.zeros( (labels.size,), dtype=np.float32)
  cdef uint32_t[:] labelct = np.zeros( (labels.size,), dtype=np.uint32)

  cdef size_t sx, sy
  sx = labels.shape[0]
  sy = labels.shape[1]

  cdef size_t x, y
  cdef uint32_t label = 0

  for x in range(sx):
    for y in range(sy):
      label = labels[x,y]
      if label == 0:
        continue

      xsum[label] += x 
      ysum[label] += y 
      labelct[label] += 1

  result = {}

  cdef float cx = wx * sx / 2
  cdef float cy = wy * sy / 2

  cdef float px, py

  for label in range(labels.size):
    if labelct[label] == 0:
      continue

    px = wx * <float>xsum[label] / <float>labelct[label]
    py = wy * <float>ysum[label] / <float>labelct[label]

    # Since we don't know which coordinate frame we 
    # are using, round toward the center of the image
    # to ensure we get the same pixel every time.
    if px - cx >= 0:
      px = px # will be truncated towards center
    else:
      px = px + wx

    if py - cy >= 0:
      py = py # will be truncated towards center
    else:
      py = py + wy

    result[label] = (<int>(px / wx), <int>(py / wy))

  return result

@cython.binding(True)
def find_border_targets(
    cnp.ndarray[float, ndim=2] dt,
    cnp.ndarray[UINT, ndim=2] cc_labels,
    float wx, float wy
  ):
  """
  Given a set of connected components that line within 
  a plane and their distance transform, return a map of
  label ID to the coordinate of its maximum distance 
  transform value. If there are multiple maxima, we 
  disambiguate based on topological criteria that are
  coordinate frame independent in order to avoid dealing
  with issues that come from the six rotated frames and
  their mirrored partners.

  The purpose of this function is to fix the edge effect
  the standard TEASAR algorithm generates and ensure that
  we can trivially join skeletons from adjacent chunks.  

  Rotating the (x,y) pairs into their appropriate frame
  is performed in the function that calls this one.

  Returns: { $SEGID: (x, y), ... }
  """
  cdef size_t sx, sy
  sx = dt.shape[0]
  sy = dt.shape[1]

  cdef size_t x, y

  mx = defaultdict(float)
  pts = {}

  cdef UINT label = 0
  cdef dict centroids = compute_centroids(cc_labels, wx, wy)
  cdef float px, py
  cdef float centx, centy

  for y in range(sy):
    for x in range(sx):
      label = cc_labels[x,y]
      if label == 0:
        continue
      elif dt[x,y] == 0:
        continue
      elif dt[x,y] > mx[label]:
        mx[label] = dt[x,y]
        pts[label] = (x,y)
      elif mx[label] == dt[x,y]:
        px, py = pts[label]
        centx, centy = centroids[label]
        pts[label] = compute_tiebreaker_maxima(
          px, py, x, y, 
          centx, centy,
          sx, sy, wx, wy
        )

  return pts

def compute_tiebreaker_maxima(
    float px, float py, 
    float x, float y, 
    float centx, float centy,
    float sx, float sy,
    float wx, float wy
  ):
  """
  compute_tiebreaker_maxima(
    float px, float py, 
    float x, float y, 
    float centx, float centy,
    float sx, float sy,
    float wx, float wy
  )

  This function breaks ties for `compute_border_targets`.

  (px,py): A previously found distance transform maxima 
  (x,y): The coordinate of the newly found maxima
  (sx,sy): The length and width of the image plane.
  (wx,wy): Weighting for anisotropy.
  (centx, centy): The centroid of the current label.

  We use following topolological criteria to achieve
  a coordinate frame-free voxel selection. We pick
  the result of the first criterion that is satisfied.

  1) Pick the voxel closest to the centroid of the label.
  2) The voxel closest to the centroid of the plane.
  3) Closest to a corner of the plane.
  4) Closest to an edge of the plane.
  5) The previous maxima.

  The worst case would be an annulus drawn around the center,
  which would result in four equally eligible pixels....

  Hopefully this won't happen too often...

  Returns: some (x, y)
  """
  cdef float cx = wx * sx / 2.0
  cdef float cy = wy * sy / 2.0

  cdef float dist1 = distsq(px,py, centx,centy, wx,wy)
  cdef float dist2 = distsq( x, y, centx,centy, wx,wy)

  if dist2 < dist1:
    return (x, y)
  elif dist1 == dist2:
    dist1 = distsq(px,py, cx,cy, wx,wy)
    dist2 = distsq( x, y, cx,cy, wx,wy)
    if dist2 < dist1:
      return (x,y)
    elif dist1 == dist2:
      dist1 = cornerness(px, py, sx, sy, wx,wy)
      dist2 = cornerness( x,  y, sx, sy, wx,wy)
      if dist2 < dist1:
        return (x, y)
      elif dist1 == dist2:
        dist1 = edgeness(px, py, sx, sy, wx,wy)
        dist2 = edgeness( x,  y, sx, sy, wx,wy)
        if dist2 < dist1:
          return (x, y)

  return (px, py)

cdef float edgeness(
    float x, float y, float sx, float sy,
    float wx, float wy
  ):
  """
  float edgeness(float x, float y, float sx, float sy)

  Nearness of (x,y) to the edge of an image of size (sx,sy).
  """
  return min(
    wx * (x - 0.5),
    wx * (sx - 0.5 - x),
    wy * (y - 0.5),
    wy * (sy - 0.5 - y)
  )

cdef float cornerness(
    float x, float y, float sx, float sy,
    float wx, float wy
  ):
  """
  float cornerness(
      float x, float y, float sx, float sy
      float wx, float wy
  )

  Nearness of (x,y) to a corner of an image of size (sx,sy).
  """
  return min( 
    distsq(x,y,-0.5,-0.5, wx, wy), 
    distsq(x,y,sx-0.5,-0.5, wx, wy),
    distsq(x,y,sx-0.5,sy-0.5, wx, wy),
    distsq(x,y,-0.5,sx-0.5, wx, wy)
  )

cdef float distsq(
    float p1x, float p1y, 
    float p2x, float p2y, 
    float wx, float wy
  ):

  p1x = wx * (p1x - p2x)
  p1y = wy * (p1y - p2y)
  return p1x * p1x + p1y * p1y 

@cython.boundscheck(False)
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.binding(True)
def roll_invalidation_cube(
    cnp.ndarray[uint8_t, cast=True, ndim=3] labels, 
    cnp.ndarray[float, ndim=3] DBF, 
    path, float scale, float const,
    anisotropy=(1,1,1),
    invalid_vertices={},
  ):
  """
  Given an anisotropic binary image, its distance transform, and a path 
  traversing the binary image, erase the voxels surrounding the path
  in a cube around each vertex. In contrast to `roll_invalidation_ball`,
  this function runs in time linear in the number of image pixels.
  """
  cdef int64_t sx, sy, sz 
  sx = labels.shape[0]
  sy = labels.shape[1]
  sz = labels.shape[2]

  cdef size_t sxy = sx * sy

  cdef float wx, wy, wz
  (wx, wy, wz) = anisotropy

  path = [ 
    coord[0] + sx * coord[1] + sxy * coord[2] 
    for coord in path if tuple(coord) not in invalid_vertices 
  ]
  path = np.array(path, dtype=np.uintp)

  cdef size_t[:] pathview = path

  cdef size_t invalidated = _roll_invalidation_cube(
    <uint8_t*>&labels[0,0,0], <float*>&DBF[0,0,0],
    sx, sy, sz, 
    wx, wy, wz,
    <size_t*>&pathview[0], path.size,
    scale, const
  )

  return invalidated, labels

@cython.boundscheck(False)
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
def find_cycle_cython(cnp.ndarray[int32_t, ndim=2] edges):
  """
  Given a graph of edges that are a single connected component,
  find a cycle via depth first search.

  Returns: list of edges in a cycle (empty list if no cycle is found)
  """
  index = defaultdict(set)
  visited = defaultdict(int)

  if edges.size == 0:
    return np.array([], dtype=np.int32)

  for e1, e2 in edges:
    index[e1].add(e2)
    index[e2].add(e1)

  cdef int root = edges[0,0]
  cdef int node = -1
  cdef int child = -1
  cdef int parent = -1
  cdef int depth = -1
  cdef int i = 0

  cdef list stack = [root]
  cdef list parents = [-1]
  cdef list depth_stack = [0]
  cdef list path = []

  while stack:
    node = stack.pop()
    parent = parents.pop()
    depth = depth_stack.pop()

    for i in range(len(path) - depth):
      path.pop()

    path.append(node)

    if visited[node] == 1:
      break

    visited[node] = 1

    for child in index[node]:
      if child != parent:
        stack.append(child)
        parents.append(node)
        depth_stack.append(depth + 1)

  if len(path) <= 1:
    return np.array([], dtype=np.int32)
  
  for i in range(len(path) - 1):
    if path[i] == node:
      break

  path = path[i:]

  if len(path) < 3:
    return np.array([], dtype=np.int32)

  return np.array(path, dtype=np.int32)

def find_avocado_fruit(
  cnp.ndarray[INTEGER, ndim=3] labels, 
  size_t cx, size_t cy, size_t cz,
  INTEGER background = 0
):
  """
  Tests to see if the current coordinate is inside 
  the nucleus of a somata that has been assigned
  to a separate label from the rest of the cell.

  Returns: (pit, fruit)
  """
  cdef size_t sx, sy, sz
  sx, sy, sz = labels.shape[:3]
  cdef size_t voxels = sx * sy * sz 

  if cx >= sx or cy >= sy or cz >= sz:
    raise ValueError(
      "<{},{},{}> must be be contained within shape <{},{},{}>".format(
        cx,cy,cz,sx,sy,sz
    ))

  cdef size_t x, y, z 
  cdef INTEGER label = labels[cx, cy, cz]
  cdef list changes = [ None ] * 6

  for x in range(cx, sx):
    if labels[x,cy,cz] == background:
      break
    elif labels[x,cy,cz] != label:
      changes[0] = labels[x,cy,cz]
      break

  for x in range(cx, 0, -1):
    if labels[x,cy,cz] == background:
      break
    elif labels[x,cy,cz] != label:
      changes[1] = labels[x,cy,cz]
      break

  for y in range(cy, sy):
    if labels[cx,y,cz] == background:
      break
    if labels[cx,y,cz] != label:
      changes[2] = labels[cx,y,cz]
      break

  for y in range(cy, 0, -1):
    if labels[cx,y,cz] == background:
      break
    if labels[cx,y,cz] != label:
      changes[3] = labels[cx,y,cz]
      break

  for z in range(cz, sz):
    if labels[cx,cy,z] == background:
      break
    if labels[cx,cy,z] != label:
      changes[4] = labels[cx,cy,z]
      break

  for z in range(cz, 0, -1):
    if labels[cx,cy,z] == background:
      break
    if labels[cx,cy,z] != label:
      changes[5] = labels[cx,cy,z]
      break

  changes = [ _ for _ in changes if _ is not None ]

  # Too little info to make a decision
  if len(changes) < 3:
    return (label, label)

  if len(changes) > 3: # if more than 3, allow one non-match
    allowed_differences = 1
  else: # allow no non-matches (we're in a corner)
    allowed_differences = 0

  uniq, cts = np.unique(changes, return_counts=True)
  candidate_fruit_index = np.argmax(cts)
  differences = len(changes) - cts[candidate_fruit_index]

  # it's not an avocado if there's lots of
  # labels surrounding the candidate "pit"
  if differences > allowed_differences:
    return (label, label)
  
  return (label, uniq[candidate_fruit_index])

class CachedTargetFinder:
  def __init__(self, mask: np.ndarray, daf: np.ndarray):
    """
    From DAF, compute a sorted list of the maximum values
    so that finding them becomes very fast.
    """
    mask_indices = np.flatnonzero(mask.ravel(order='F'))
    if mask.size < np.iinfo(np.uint32).max:
      mask_indices = mask_indices.astype(np.uint32, copy=False)
    daf_sort = np.argsort(daf.ravel(order='F')[mask_indices])
    daf_sort = np.flip(daf_sort)
    self.daf_indices = mask_indices[daf_sort]

  def find_target(self, mask: np.ndarray):
    """
    Find the coordinate of a voxel corresponding 
    the maximum map value.

    Returns: (x, y, z)
    """
    first_positive_index = self.first_label_indexed(
      mask.ravel(order='F'), self.daf_indices
    )
    if first_positive_index is None:
      self.daf_indices = self.daf_indices[self.daf_indices.size:]  # Clear it.
      return None

    # This tells us mask positions daf_indices[0:first_positive_index] are now
    # zeroed out. We assume that this is permanent, so we don't need to search
    # those positions again next time.
    self.daf_indices = self.daf_indices[first_positive_index:]

    return np.unravel_index(self.daf_indices[0], mask.shape, order='F')

  @cython.boundscheck(False)
  @cython.wraparound(False)  # turn off negative index wrapping for entire function
  @cython.nonecheck(False)
  def first_label_indexed(self, uint8_t[:] labels not None, INTEGER[:] indices not None):
    """
    Returns: first i for which labels[indices[i]] is non-zero.
    """
    cdef size_t length = indices.size
    cdef size_t i = 0
    cdef INTEGER label_index

    for i in range(length):
      label_index = indices[i]
      if labels[label_index]:
        return i

    return None  

def extract_edges_from_binary_image(uint8_t[:,:,:] binimg, int connectivity = 26):
  cdef uint64_t sx, sy, sz
  sx, sy, sz = tuple(binimg.shape)[:3]

  cdef uint64_t sxy = sx * sy

  binimg = np.asfortranarray(binimg)
  cdef unordered_set[cpp_pair[uint64_t,uint64_t], pair_hash] edges = _extract_edges_from_binary_image(
    &binimg[0,0,0], 
    sx, sy, sz, 
    connectivity
  )

  numbering = {}
  cdef int64_t i = 0
  for edge in edges:
    for v in (edge.first, edge.second):
      if v not in numbering:
        numbering[v] = i
        i += 1

  inumbering = { v:k for k,v in numbering.items() }
  vertices = []

  cdef uint64_t loc, x, y, z
  for i in range(len(inumbering)):
    loc = inumbering[i]
    z = loc // sxy
    y = (loc - z * sxy) // sx
    x = loc - z * sxy - y * sx
    vertices.append((x,y,z))

  int_edges = []
  for v1,v2 in edges:
    int_edges.append((numbering[v1], numbering[v2]))

  vertices = np.array(vertices, dtype=np.uint32)
  int_edges = np.array(int_edges, dtype=np.uint32)

  return (vertices, int_edges)






from typing import Dict, Union, List, Tuple, Optional

from collections import defaultdict
import copy
import os

import numpy as np
import numpy.typing as npt
import scipy.ndimage
from tqdm import tqdm

from osteoid import Skeleton, Bbox, Vec

import kimimaro.skeletontricks

import cc3d
from crackle import CrackleArray
import dijkstra3d
import fastremap
import fill_voids
import xs3d

XS_PROP = {
  "id": "cross_sectional_area",
  "data_type": "float32",
  "num_components": 1,
}

XS_CONTACT_PROP = {
  "id": "cross_sectional_area_contacts",
  "data_type": "uint8",
  "num_components": 1,  
}

def toabs(path):
  path = os.path.expanduser(path)
  return os.path.abspath(path)

def mkdir(path):
  path = toabs(path)

  try:
    if path != '' and not os.path.exists(path):
      os.makedirs(path)
  except OSError as e:
    if e.errno == 17: # File Exists
      time.sleep(0.1)
      return mkdir(path)
    else:
      raise

  return path

def extract_skeleton_from_binary_image(image):
  verts, edges = kimimaro.skeletontricks.extract_edges_from_binary_image(image)
  return Skeleton(verts, edges)

def compute_cc_labels(all_labels, voxel_graph = None):
  if isinstance(all_labels, CrackleArray):
    if voxel_graph is not None:
      all_labels = all_labels[:]
    else:
      return all_labels.connected_components(
        connectivity=26,
        memory_target=int(500e6), 
        return_mapping=True,
      )

  tmp_labels = all_labels
  if np.dtype(all_labels.dtype).itemsize > 1:
    tmp_labels, remapping = fastremap.renumber(all_labels, in_place=False)

  if voxel_graph is not None:
    cc_labels = cc3d.color_connectivity_graph(voxel_graph, connectivity=26)
    cc_labels *= all_labels > 0
  else:
    cc_labels = cc3d.connected_components(tmp_labels)
  
  cc_labels = fastremap.refit(cc_labels)

  del tmp_labels
  remapping = kimimaro.skeletontricks.get_mapping(all_labels, cc_labels) 
  return cc_labels, remapping

def find_objects(labels):
  """  
  scipy.ndimage.find_objects performs about 7-8x faster on C 
  ordered arrays, so we just do it that way and convert
  the results if it's in F order.
  """
  if isinstance(labels, CrackleArray):
    bbxes = labels.bounding_boxes()
    bbxes.pop(0)
    result = list(bbxes.items())
    result.sort(key=lambda x: x[0])
    return [ x[1] for x in result ]

  if labels.flags['C_CONTIGUOUS']:
    return scipy.ndimage.find_objects(labels)
  else:
    all_slices = scipy.ndimage.find_objects(labels.T)
    return [ (slcs and slcs[::-1]) for slcs in all_slices ]    

def add_property(skel, prop):
  needs_prop = True
  for skel_prop in skel.extra_attributes:
    if skel_prop["id"] == prop["id"]:
      needs_prop = False
      break

  if needs_prop:
    skel.extra_attributes.append(prop)

def shape_iterator(all_labels, skeletons, fill_holes, in_place, progress, fn):
  iterator = skeletons
  if type(skeletons) == dict:
    iterator = skeletons.values()
    total = len(skeletons)
  elif hasattr(skeletons, "vertices"):
    iterator = [ skeletons ]
    total = 1
  else:
    total = len(skeletons)

  if all_labels.dtype == bool:
    remapping = { True: 1, False: 0, 1:1, 0:0 }
  else:
    all_labels, remapping = fastremap.renumber(all_labels, in_place=in_place)

  all_slices = find_objects(all_labels)

  with tqdm(iterator, desc="Labels", disable=(not progress), total=total) as pbar:
    for skel in pbar:
      if all_labels.dtype == bool:
        label = 1
      else:
        label = skel.id

      pbar.set_postfix(label=str(label))

      if label == 0:
        continue

      if label not in remapping:
        continue

      label = remapping[label]
      slices = all_slices[label - 1]
      if slices is None:
        continue

      roi = Bbox.from_slices(slices)
      if roi.volume() <= 1:
        continue

      roi.grow(1)
      roi.minpt = Vec.clamp(roi.minpt, Vec(0,0,0), roi.maxpt)
      slices = roi.to_slices()

      binimg = np.asfortranarray(all_labels[slices] == label)
      if fill_holes:
        binimg = fill_voids.fill(binimg, in_place=True)

      fn(skel, binimg, roi)

  return iterator

def cross_sectional_area_single(
  binimg:npt.NDArray[np.bool_], 
  skel:Skeleton, 
  roi:Optional[Bbox] = None,
  anisotropy:npt.NDArray[np.float32] = np.array([1,1,1], dtype=np.float32),
  smoothing_window:int = 1,
  progress:bool = False,
  in_place:bool = False,
  repair_contacts:bool = False,
  visualize_section_planes:bool = False,
  step:int = 1,
) -> Skeleton:
  """
  Analyze the cross sectional area for a single skeleton given 
  an overlapping binary image. For many skeletons at once, 
  use cross_sectional_area which may be faster.

  When the smoothing_window is >1, these plane normal 
  vectors will be smoothed with a rolling average. This
  is useful since there can be high frequency
  oscillations in the skeleton.

  This function will add the following attributes to
  each skeleton provided.

  skel.cross_sectional_area: float32 array of cross 
    sectional area per a vertex.

  skel.cross_sectional_area_contacts: uint8 array
    where non-zero entries indicate that the image
    border was contacted during the cross section
    computation, indicating a possible underestimate.

    The first six bits are a bitfield xxyyzz that
    tell you which image faces were touched and
    alternate from low (0) to high (size-1).

  repair_contacts: When True, only examine vertices
    that have a nonzero value for 
    skel.cross_sectional_area_contacts. This is intended
    to be used as a second pass after widening the image.

  visualize_section_planes: For debugging, paint section planes
    and display them using microviewer.

  step: when > 1, skip (step-1) vertices. This can be used to
    go faster. These days, evaluating a single vertex takes 
    between a few hundred microseconds to a few thousand microseconds.
      example calculation: 
      1 msec x 100,000 vertices = 100 sec
      A neuron I recently examined had over 300,000 vertices across 
      the entire dataset.
      Kimimaro's benchmark task produced 622,293 vertices over 1667 objects 
      using reasonable parameters and took a little over 4 minutes on an M3 
      processor (or about 2.5 msec/vertex). The most expensive shape was the soma.
  """
  assert step > 0
  assert smoothing_window > 0

  cross_sections = None
  if visualize_section_planes:
    cross_sections = np.zeros(binimg.shape, dtype=np.uint32, order="F")

  if skel.space == "physical":
    all_verts = (skel.vertices / anisotropy).round().astype(int)
  else:
    all_verts = np.copy(skel.vertices)

  if roi is not None:
    all_verts -= roi.minpt

  mapping = { tuple(v): i for i, v in enumerate(all_verts) }

  visited = np.zeros([ all_verts.shape[0] ], dtype=bool)

  if repair_contacts:
    areas = skel.cross_sectional_area
    contacts = skel.cross_sectional_area_contacts
  else:
    areas = np.zeros([all_verts.shape[0]], dtype=np.float32)
    contacts = np.zeros([all_verts.shape[0]], dtype=np.uint8)

  branch_pts = set(skel.branches())
  branch_pt_vals = defaultdict(list)

  paths = skel.paths()

  normal = np.array([1,0,0], dtype=np.float32)

  shape = np.array(binimg.shape)

  try:
    xs3d.set_shape(binimg)
    
    for path in tqdm(paths, disable=(not progress), desc="Cross Section Analysis Paths"):
      if skel.space == "physical":
        path = (path / anisotropy).round().astype(int)
      if roi is not None:
        path -= roi.minpt

      normals = (path[1:] - path[:-1]).astype(np.float32)
      normals = np.concatenate([ normals, [normals[-1]] ])

      # Running the filter in the forward and then backwards
      # direction eliminates phase shift.
      normals = moving_average(normals, smoothing_window)
      normals = moving_average(normals[::-1], smoothing_window)[::-1]

      normals /= np.linalg.norm(normals, axis=1, keepdims=True)   

      end_i = len(path) - 1
      ct = 0

      for i, vert in enumerate(path):
        ct += 1

        if ct < step and not (i == 0 or i == end_i):
          continue
        elif ct == step:
          ct = 0

        if ( 
             (vert[0] < 0) 
          or (vert[0] >= shape[0])
          or (vert[1] < 0) 
          or (vert[1] >= shape[1])
          or (vert[2] < 0) 
          or (vert[2] >= shape[2])
        ):
          continue

        idx = mapping[tuple(vert)]
        normal = normals[i]

        if (
          areas[idx] == 0 
          or (idx in branch_pts) 
          or (repair_contacts and contacts[idx] > 0 and not visited[idx])
        ):
          visited[idx] = True
          areas[idx], contact = xs3d.cross_sectional_area(
            binimg, vert, 
            normal, anisotropy,
            return_contact=True,
            use_persistent_data=True,
          )
          if repair_contacts:
            contacts[idx] = contact
          else:
            contacts[idx] |= contact # accumulate for branch points
          if idx in branch_pts:
            branch_pt_vals[idx].append(areas[idx])
          if visualize_section_planes:
            img = xs3d.cross_section(
              binimg, vert, 
              normal, anisotropy,
            )
            cross_sections[img > 0] = idx
  finally:
    xs3d.clear_shape()

  if visualize_section_planes:
    import microviewer
    microviewer.view(cross_sections, seg=True)

  for idx, vals in branch_pt_vals.items():
    areas[idx] = sum(vals) / len(vals)

  skel.cross_sectional_area = areas
  skel.cross_sectional_area_contacts = contacts

  add_property(skel, XS_PROP)
  add_property(skel, XS_CONTACT_PROP)

  return skel

def cross_sectional_area(
  all_labels:np.ndarray, 
  skeletons:Union[Dict[int,Skeleton],List[Skeleton],Skeleton],
  anisotropy:np.ndarray = np.array([1,1,1], dtype=np.float32),
  smoothing_window:int = 1,
  progress:bool = False,
  in_place:bool = False,
  fill_holes:bool = False,
  repair_contacts:bool = False,
  visualize_section_planes:bool = False,
  step:int = 1,
) -> Union[Dict[int,Skeleton],List[Skeleton],Skeleton]:
  """
  Given a set of skeletons, find the cross sectional area
  for each vertex indicated by the sectioning plane
  defined by the vector pointing to the next vertex.

  When the smoothing_window is >1, these plane normal 
  vectors will be smoothed with a rolling average. This
  is useful since there can be high frequency
  oscillations in the skeleton.

  This function will add the following attributes to
  each skeleton provided.

  skel.cross_sectional_area: float32 array of cross 
    sectional area per a vertex.

  skel.cross_sectional_area_contacts: uint8 array
    where non-zero entries indicate that the image
    border was contacted during the cross section
    computation, indicating a possible underestimate.

    The first six bits are a bitfield xxyyzz that
    tell you which image faces were touched and
    alternate from low (0) to high (size-1).

  repair_contacts: When True, only examine vertices
    that have a nonzero value for 
    skel.cross_sectional_area_contacts. This is intended
    to be used as a second pass after widening the image.

  visualize_section_planes: For debugging, paint section planes
    and display them using microviewer.

  step: when > 1, skip (step-1) vertices. This can be used to
    go faster. These days, evaluating a single vertex takes 
    between a few hundred microseconds to a few thousand microseconds.
      example calculation: 
      1 msec x 100,000 vertices = 100 sec
      A neuron I recently examined had over 300,000 vertices across 
      the entire dataset.
      Kimimaro's benchmark task produced 622,293 vertices over 1667 objects 
      using reasonable parameters and took a little over 4 minutes on an M3 
      processor (or about 2.5 msec/vertex). The most expensive shape was the soma.
  """
  assert step > 0
  assert smoothing_window > 0

  def cross_sectional_area_helper(skel, binimg, roi):
    cross_sections = None
    if visualize_section_planes:
      cross_sections = np.zeros(binimg.shape, dtype=np.uint32, order="F")

    if skel.space == "physical":
      all_verts = (skel.vertices / anisotropy).round().astype(int)
    else:
      all_verts = np.copy(skel.vertices)

    all_verts -= roi.minpt

    mapping = { tuple(v): i for i, v in enumerate(all_verts) }

    visited = np.zeros([ all_verts.shape[0] ], dtype=bool)

    if repair_contacts:
      areas = skel.cross_sectional_area
      contacts = skel.cross_sectional_area_contacts
    else:
      areas = np.zeros([all_verts.shape[0]], dtype=np.float32)
      contacts = np.zeros([all_verts.shape[0]], dtype=np.uint8)

    branch_pts = set(skel.branches())
    branch_pt_vals = defaultdict(list)

    paths = skel.paths()

    normal = np.array([1,0,0], dtype=np.float32)

    shape = np.array(binimg.shape)

    for path in paths:
      if skel.space == "physical":
        path = (path / anisotropy).round().astype(int)
      path -= roi.minpt

      normals = (path[1:] - path[:-1]).astype(np.float32)
      normals = np.concatenate([ normals, [normals[-1]] ])

      # Running the filter in the forward and then backwards
      # direction eliminates phase shift.
      normals = moving_average(normals, smoothing_window)
      normals = moving_average(normals[::-1], smoothing_window)[::-1]

      normals /= np.linalg.norm(normals, axis=1, keepdims=True)   

      end_i = len(path) - 1
      ct = 0

      for i, vert in enumerate(path):
        ct += 1

        if ct < step and not (i == 0 or i == end_i):
          continue
        elif ct == step:
          ct = 0

        if ( 
             (vert[0] < 0) 
          or (vert[0] >= shape[0])
          or (vert[1] < 0) 
          or (vert[1] >= shape[1])
          or (vert[2] < 0) 
          or (vert[2] >= shape[2])
        ):
          continue

        idx = mapping[tuple(vert)]
        normal = normals[i]

        if (
          areas[idx] == 0 
          or (idx in branch_pts) 
          or (repair_contacts and contacts[idx] > 0 and not visited[idx])
        ):
          visited[idx] = True
          areas[idx], contact = xs3d.cross_sectional_area(
            binimg, vert, 
            normal, anisotropy,
            return_contact=True,
            use_persistent_data=True,
          )
          if repair_contacts:
            contacts[idx] = contact
          else:
            contacts[idx] |= contact # accumulate for branch points
          if idx in branch_pts:
            branch_pt_vals[idx].append(areas[idx])
          if visualize_section_planes:
            img = xs3d.cross_section(
              binimg, vert, 
              normal, anisotropy,
            )
            cross_sections[img > 0] = idx

    if visualize_section_planes:
      import microviewer
      microviewer.view(cross_sections, seg=True)

    for idx, vals in branch_pt_vals.items():
      areas[idx] = sum(vals) / len(vals)

    skel.cross_sectional_area = areas
    skel.cross_sectional_area_contacts = contacts

  try:
    xs3d.set_shape(all_labels)
    if isinstance(all_labels, CrackleArray):
      bboxes = all_labels.bounding_boxes()
      iterator = tqdm(
        all_labels.each(crop=True, labels=list(skeletons.keys())),
        disable=(not progress),
        desc="Cross Section Analysis Paths"
      )
      for label, binimg in iterator:
        slc = Bbox.from_slices(bboxes[label])
        cross_sectional_area_helper(skeletons[label], binimg, slc)
    else:
      shape_iterator(
        all_labels, skeletons, 
        fill_holes, in_place, progress, 
        cross_sectional_area_helper
      )
  finally:
    xs3d.clear_shape()

  if hasattr(skeletons, "vertices"):
    skelitr = [ skeletons ]
  elif isinstance(skeletons, dict):
    skelitr = skeletons.values()
  else:
    skelitr = iter(skeletons)

  for skel in skelitr:
    add_property(skel, XS_PROP)
    add_property(skel, XS_CONTACT_PROP)

    if not hasattr(skel, "cross_sectional_area"):
      skel.cross_sectional_area = np.full(len(skel.vertices), -1, dtype=np.float32, order="F")
    if not hasattr(skel, "cross_sectional_area_contacts"):
      skel.cross_sectional_area_contacts = np.zeros(len(skel.vertices), dtype=np.uint8, order="F")

  return skeletons

def oversegment(
  all_labels:np.ndarray, 
  skeletons:Union[Dict[int,Skeleton],List[Skeleton],Skeleton],
  anisotropy:np.ndarray = np.array([1,1,1], dtype=np.float32),
  progress:bool = False,
  fill_holes:bool = False,
  in_place:bool = False,
  downsample:int = 0,
) -> Tuple[np.ndarray, Union[Dict[int,Skeleton],List[Skeleton],Skeleton]]:
  """
  Use skeletons to create an oversegmentation of a pre-existing set
  of labels. This is useful for proofreading systems that work by merging
  labels.

  For each skeleton, get the feature map from its euclidean distance
  field. The final image is the composite of all these feature maps
  numbered from 1.

  Each skeleton will have a new property skel.segments that associates
  a label to each vertex.
  """
  prop = {
    "id": "segments",
    "data_type": "uint64",
    "num_components": 1,
  }

  skeletons = copy.deepcopy(skeletons)

  # Initialize segments attribute for all skeletons
  if hasattr(skeletons, "vertices"):
    skeleton_list = [skeletons]
  elif isinstance(skeletons, dict):
    skeleton_list = list(skeletons.values())
  else:
    skeleton_list = skeletons
    
  for skel in skeleton_list:
    if not hasattr(skel, 'segments'):
      skel.segments = np.zeros(len(skel.vertices), dtype=np.uint64)

  all_features = np.zeros(all_labels.shape, dtype=np.uint64, order="F")
  next_label = 0

  def oversegment_helper(skel, binimg, roi):
    nonlocal next_label
    nonlocal all_features

    segment_skel = skel
    if downsample > 0:
      segment_skel = skel.downsample(downsample)

    vertices = (segment_skel.vertices / anisotropy).round().astype(int)
    vertices -= roi.minpt

    field, feature_map = dijkstra3d.euclidean_distance_field(
      binimg, vertices, 
      anisotropy=anisotropy, 
      return_feature_map=True
    )
    del field

    add_property(skel, prop)

    vertices = (skel.vertices / anisotropy).round().astype(int)
    vertices -= roi.minpt

    feature_map[binimg] += next_label
    skel.segments = feature_map[vertices[:,0], vertices[:,1], vertices[:,2]]
    next_label += vertices.shape[0]
    all_features[roi.to_slices()] += feature_map

  # iterator is an iterable list of skeletons, not the shape iterator
  iterator = shape_iterator(
    all_labels, skeletons, fill_holes, in_place, progress, 
    oversegment_helper
  )

  all_features, mapping = fastremap.renumber(all_features)
  for skel in iterator:
    skel.segments = fastremap.remap(skel.segments, mapping, in_place=True)

  return all_features, skeletons

# From SO: https://stackoverflow.com/questions/14313510/how-to-calculate-rolling-moving-average-using-python-numpy-scipy
def moving_average(a:np.ndarray, n:int, mode:str = "symmetric") -> np.ndarray:
  if n <= 0:
    raise ValueError(f"Window size ({n}), must be >= 1.")
  elif n == 1:
    return a

  if len(a) == 0:
    return a

  if a.ndim == 2:
    a = np.pad(a, [[n, n],[0,0]], mode=mode)
  else:
    a = np.pad(a, [n, n], mode=mode)

  ret = np.cumsum(a, dtype=float, axis=0)
  ret = (ret[n:] - ret[:-n])[:-n]
  ret /= float(n)
  return ret




#!/usr/bin/env python
import os
import setuptools
import sys

class NumpyImport:
  def __repr__(self):
    import numpy as np

    return np.get_include()

  __fspath__ = __repr__

extra_compile_args = []
if sys.platform == 'win32':
  extra_compile_args += [
    '/std:c++17', '/O2'
  ]
else:
  extra_compile_args += [
    '-std=c++17', '-O3'
  ]

if sys.platform == 'darwin':
  extra_compile_args += [ '-stdlib=libc++', '-mmacosx-version-min=10.9' ]

setuptools.setup(
    ext_modules=[
      setuptools.Extension(
        'kimimaro.skeletontricks',
        sources=[ './ext/skeletontricks/skeletontricks.pyx' ],
        language='c++',
        include_dirs=[ str(NumpyImport()) ],
        extra_compile_args=extra_compile_args,
      ),
    ],
)